"""Writer."""

from __future__ import annotations

import typing as ty
from contextlib import contextmanager
from pathlib import Path

import cv2
import numpy as np
import SimpleITK as sitk
from koyo.decorators import retry
from koyo.timer import MeasureTimer
from loguru import logger
from tifffile import TiffWriter
from tqdm import tqdm

from image2image_io.readers.utilities import get_pyramid_info, prepare_ome_xml_str

if ty.TYPE_CHECKING:
    from image2image_io.readers._base_reader import BaseReader


class Transformer(ty.Protocol):
    """Transformer protocol to transform image."""

    output_size: tuple[int, int]
    output_spacing: tuple[float, float]

    def __call__(self, image: sitk.Image) -> sitk.Image:
        """Transform image."""
        ...


class OmeTiffWriter:
    """OME-TIFF writer class.

    Attributes
    ----------
    x_size: int
        Size of the output image after transformation in x
    y_size: int
        Size of the output image after transformation in y
    tile_size: int
        Size of tiles to be written
    pyr_levels: list of tuples of int:
        Size of down-sampled images in pyramid
    n_pyr_levels: int
        Number of downsamples in pyramid
    PhysicalSizeY: float
        physical size of image in micron for OME-TIFF in Y
    PhysicalSizeX: float
        physical size of image in micron for OME-TIFF in X
    subifds: int
        Number of sub-resolutions for pyramidal OME-TIFF
    compression: str
        tifffile string to pass to compression argument, defaults to "deflate" for minisblack
        and "jpeg" for RGB type images
    """

    x_size: int | None = None
    y_size: int | None = None
    tile_size: int = 512
    pyr_levels: list[tuple[int, int]]
    n_pyr_levels: int
    PhysicalSizeY: int | float
    PhysicalSizeX: int | float
    subifds: int | None = None
    compression: str | None = "default"

    def __init__(
        self,
        reader: BaseReader,
        transformer: Transformer | None = None,
        crop_mask: np.ndarray | None = None,
        crop_bbox: tuple[int, int, int, int] | None = None,
    ):
        """Class for managing writing images to OME-TIFF.

        Parameters
        ----------
        reader: BaseReader
            MergeRegImage to be transformed
        transformer: Transformer or None
            Registration transformation sequences for each wsireg image to be merged
        crop_mask : np.ndarray or None
            Crop mask to apply to images before writing.
        crop_bbox : tuple of ints or None
            Bounding box to crop images to before writing. Values should be x, y, width, height.
        """
        self.reader = reader
        self.transformer = transformer
        self.crop_mask = crop_mask
        self.crop_bbox = self._check_bbox(crop_bbox)
        if self.crop_mask is not None and self.crop_bbox is not None:
            raise ValueError("Cannot supply both crop_mask and crop_bbox")

    def _check_bbox(self, crop_bbox: tuple[int, int, int, int] | None) -> tuple[int, int, int, int] | None:
        """Check bbox."""
        if crop_bbox is None:
            return None
        x, y, width, height = crop_bbox
        if x < 0:
            x = 0
        if y < 0:
            y = 0
        if self.transformer is not None:
            image_shape = self.transformer.output_size[::-1]
        else:
            image_shape = self.reader.image_shape
        if x + width > image_shape[1]:
            width = image_shape[1] - x
        if y + height > image_shape[0]:
            height = image_shape[0] - y
        return x, y, width, height

    def _prepare_image_info(
        self,
        image_name: str,
        write_pyramid: bool = True,
        tile_size: int = 512,
        compression: str | None = "default",
        as_uint8: bool | None = None,
        channel_ids: list[int | tuple[int, ...]] | None = None,
        channel_names: list[str] | None = None,
        image_shape: tuple[int, int] | None = None,
        n_channels: int | None = None,
        resolution: float | None = None,
        dtype: np.dtype | None = None,
        is_rgb: bool | None = None,
    ) -> None:
        """Get image info and OME-XML."""
        if n_channels is None:
            n_channels = self.reader.n_channels
        if resolution is None:
            resolution = self.reader.resolution
        if dtype is None:
            dtype = self.reader.dtype
        if is_rgb is None:
            is_rgb = self.reader.is_rgb

        if self.transformer:
            self.x_size, self.y_size = self.transformer.output_size
        elif self.crop_mask is not None:
            self.y_size, self.x_size = self.crop_mask.shape
        elif image_shape is not None and len(image_shape) == 2:
            self.y_size, self.x_size = image_shape
        else:
            self.y_size, self.x_size = self.reader.image_shape

        if self.crop_bbox is not None:
            _, _, self.x_size, self.y_size = self.crop_bbox
        if not write_pyramid:
            tile_size = 0

        if tile_size:
            # protect against too large tile size
            while self.y_size / tile_size <= 1 or self.x_size / tile_size <= 1:
                tile_size = tile_size // 2
            if tile_size < 256:
                tile_size = 256
            if self.x_size < tile_size or self.y_size < tile_size:
                tile_size = 0
        self.tile_size = tile_size
        write_pyramid = self.tile_size > 0

        if self.tile_size:
            self.pyr_levels, _ = get_pyramid_info(self.y_size, self.x_size, n_channels, self.tile_size)
        else:
            self.pyr_levels = [(self.y_size, self.x_size)]
        self.n_pyr_levels = len(self.pyr_levels)
        write_pyramid = write_pyramid and self.n_pyr_levels > 1

        if self.transformer:
            self.PhysicalSizeX, self.PhysicalSizeY = self.transformer.output_spacing
        else:
            self.PhysicalSizeX = self.PhysicalSizeY = resolution

        is_merging = False
        if channel_ids:
            for channel_id in channel_ids:
                if not isinstance(channel_id, int):
                    is_merging = True

        if as_uint8 is None:
            as_uint8 = dtype == np.uint8
        dtype: np.dtype = np.uint8 if as_uint8 else dtype  # type: ignore[assignment,no-redef]
        if channel_names is None:
            channel_names = self.reader.channel_names
            logger.trace(f"Channel names were undefined - using {channel_names}")
        if channel_names and len(channel_names) > n_channels and not is_merging:
            channel_names = self.reader.channel_names
            logger.warning(f"Channel names were too long - using {channel_names}")
        if channel_ids is not None and not is_merging:
            if len(channel_ids) != len(channel_names):
                channel_names = [channel_names[i] for i in channel_ids]
        if len(channel_names) != len(channel_ids):  # type: ignore[arg-type]
            raise ValueError("The number of channel ids and channel names does not match.")

        self.omexml = prepare_ome_xml_str(
            self.y_size,
            self.x_size,
            len(channel_names),
            dtype,
            is_rgb,
            PhysicalSizeX=self.PhysicalSizeX,
            PhysicalSizeY=self.PhysicalSizeY,
            PhysicalSizeXUnit="µm",
            PhysicalSizeYUnit="µm",
            Name=image_name,
            Channel=None if is_rgb else {"Name": channel_names},
        )

        self.subifds = self.n_pyr_levels - 1 if write_pyramid is True else None

        if compression == "default":
            self.compression = "jpeg" if is_rgb else "deflate"
        else:
            self.compression = compression
        logger.info(f"Saving using {self.compression} compression. Datatype: {dtype} (as_uint8={as_uint8})")

    def _prepare_tiff(
        self,
        name: str,
        output_dir: Path | str = "",
        write_pyramid: bool = True,
        tile_size: int = 512,
        compression: str | None = "default",
        as_uint8: bool | None = None,
        channel_ids: list[int | tuple[int, ...]] | None = None,
        channel_names: list[str] | None = None,
        overwrite: bool = False,
        image_shape: tuple[int, int] | None = None,
        n_channels: int | None = None,
        resolution: float | None = None,
        dtype: np.dtype | None = None,
        is_rgb: bool | None = None,
        ome_name: str | None = None,
    ) -> tuple[Path | None, Path, dict[str, ty.Any] | None, list[int] | None, list[str] | None, bool | None, bool]:
        """Prepare all the necessary information to write a TIFF file."""
        name = name.replace(".ome", "").replace(".tiff", "").replace(".tif", "")
        output_file_name = Path(output_dir) / f"{name}.ome.tiff"
        tmp_output_file_name = output_file_name.parent / f"{name}.ome.tiff.tmp"
        logger.info(f"Saving to '{output_file_name}'")
        logger.info(f"Using reader: {self.reader}")
        logger.info(f"Using transformer: {self.transformer}")

        if output_file_name.exists():
            if not overwrite:
                logger.warning(f"File {output_file_name} already exists, skipping...")
                return None, output_file_name, None, None, None, as_uint8, write_pyramid
            try:
                output_file_name.unlink()
            except (PermissionError, FileNotFoundError, Exception) as e:
                raise PermissionError(f"Could not remove file {output_file_name} - {e}")

        if channel_ids and channel_names:
            if len(channel_ids) != len(channel_names):
                raise ValueError("The number of channel ids and channel names must match when being specified.")

        if channel_ids is None:
            channel_ids = list(range(self.reader.n_channels))
        if channel_names is None:
            channel_names = self.reader.channel_names

        if np.max(self.reader.image_shape) < 256:
            write_pyramid = False

        self._prepare_image_info(
            ome_name or name,
            write_pyramid=write_pyramid,
            tile_size=tile_size,
            compression=compression,
            as_uint8=as_uint8,
            channel_ids=channel_ids,
            channel_names=channel_names,
            image_shape=image_shape,
            n_channels=n_channels,
            resolution=resolution,
            dtype=dtype,
            is_rgb=is_rgb,
        )
        if as_uint8 is None:
            as_uint8 = self.reader.dtype == np.uint8
        if as_uint8:
            logger.trace(f"Writing image data in 0-255 range as uint8. Data type: {self.reader.dtype}")

        # some info about channels
        logger.trace(f"Writing channels ids: {channel_ids}")
        logger.trace(f"Writing channels: {channel_names}")

        if self.reader.is_rgb:
            options = {
                "tile": (self.tile_size, self.tile_size),
                "compression": self.compression,
                "photometric": "rgb",
                "metadata": None,
            }
        else:
            options = {
                "tile": (self.tile_size, self.tile_size),
                "compression": self.compression,
                "photometric": "minisblack",
                "metadata": None,
            }
        if self.tile_size == 0 or np.max(self.reader.image_shape) < self.tile_size:
            options.pop("tile")

        logger.trace(f"TIFF options: {options}")
        logger.trace(f"Pyramid levels: {self.pyr_levels} ({self.n_pyr_levels})")
        return tmp_output_file_name, output_file_name, options, channel_ids, channel_names, as_uint8, write_pyramid

    def write(
        self,
        name: str,
        output_dir: Path,
        tile_size: int = 512,
        as_uint8: bool | None = None,
        channel_ids: list[int | tuple[int, ...]] | None = None,
        channel_names: list[str] | None = None,
        overwrite: bool = False,
    ) -> Path | None:
        """Write image."""
        return self.write_image_by_plane(
            name,
            output_dir,
            tile_size=tile_size,
            channel_ids=channel_ids,
            as_uint8=as_uint8,
            channel_names=channel_names,
            overwrite=overwrite,
        )

    def _processed_image_yield(
        self, channel_ids: list[int | tuple[int, ...]], channel_names: list[str], as_uint8: bool | None = None
    ) -> ty.Generator[tuple[str, int, np.ndarray], None, None]:
        reader = self.reader
        for index, channel_index in enumerate(
            tqdm(
                channel_ids,
                desc="Writing channels..." if not reader.is_rgb else "Collecting channels...",
            )
        ):
            channel_name = channel_names[index]
            if channel_index not in channel_ids:
                logger.trace(f"Skipping channel {channel_index}")
                continue

            # if channel_index is int, then it's simple, otherwise, let's get maximum intensity projection
            if isinstance(channel_index, int):
                image: sitk.Image = np.squeeze(
                    reader.get_channel(channel_index, split_rgb=True),  # type: ignore[assignment]
                )
            else:
                images: list[sitk.Image] = [np.squeeze(reader.get_channel(i, split_rgb=True)) for i in channel_index]
                images = np.dstack(images)  # type: ignore[assignment]
                image = np.max(images, axis=2)  # type: ignore[assignment]
                logger.trace(f"Max intensity projection for channel {channel_name} ({channel_index})")

            image = self._process_image(image, resolution=reader.resolution, as_uint8=as_uint8)
            yield channel_name, channel_index, image

    def _process_image(
        self, image: sitk.Image | np.ndarray, resolution: float, as_uint8: bool | None = None
    ) -> np.ndarray:
        # check whether we actually need to do any pre-processing
        if self.transformer or as_uint8:
            if hasattr(image, "compute"):  # convert dask array to numpy
                image = image.compute()
            image = sitk.GetImageFromArray(image)  # type: ignore[arg-type]
            image.SetSpacing((resolution, resolution))  # type: ignore[no-untyped-call]

            # transform
            if self.transformer and callable(self.transformer):
                with MeasureTimer() as timer:
                    image = self.transformer(image)
                logger.trace(
                    f"Transformed image shape: {image.GetSize()[::-1]} ({image.GetSpacing()})in {timer()}",
                )

            # change dtype
            if as_uint8 and image.GetPixelID() != sitk.sitkUInt8:  # type: ignore[no-untyped-call]
                image = sitk.RescaleIntensity(image, 0, 255)  # type: ignore[no-untyped-call]
                image = sitk.Cast(image, sitk.sitkUInt8)  # type: ignore[no-untyped-call]

            # convert to array if necessary
            if isinstance(image, sitk.Image):
                image: np.ndarray = sitk.GetArrayFromImage(image)  # type: ignore[no-redef]

        # ensure that the image is a numpy array and not e.g. dask/zarr array
        image = np.asarray(image)  # type: ignore[assignment]

        # apply crop mask
        if self.crop_mask is not None:
            image = self.crop_mask * image
        elif self.crop_bbox is not None:
            x, y, width, height = self.crop_bbox
            image = image[y : y + height, x : x + width]
        return image

    def _convert_image_to_pyramid(
        self, image: np.ndarray, interpolation: ty.Any = cv2.INTER_LINEAR, pyr_levels: list[int] | None = None
    ) -> ty.Generator[tuple[int, np.ndarray]]:
        """Convert image to pyramid and yield the pyramid levels."""
        if pyr_levels is None:
            pyr_levels = list(range(1, self.n_pyr_levels))
        for pyramid_index in pyr_levels:
            resize_shape = (self.pyr_levels[pyramid_index][0], self.pyr_levels[pyramid_index][1])
            yield pyramid_index, cv2.resize(image, resize_shape, interpolation)

    def _write_image_rgb(
        self,
        name: str,
        output_dir: Path | str = "",
        write_pyramid: bool = True,
        tile_size: int = 512,
        compression: str | None = "default",
        as_uint8: bool | None = None,
        channel_ids: list[int | tuple[int, ...]] | None = None,
        channel_names: list[str] | None = None,
        overwrite: bool = False,
        ome_name: str | None = None,
    ) -> Path | None:
        # make sure user did not provide filename with OME-TIFF
        (
            tmp_output_file_name,
            output_file_name,
            options,
            channel_ids,
            channel_names,
            as_uint8,
            write_pyramid,
        ) = self._prepare_tiff(
            name=name,
            output_dir=output_dir,
            write_pyramid=write_pyramid,
            tile_size=tile_size,
            compression=compression,
            as_uint8=as_uint8,
            channel_ids=channel_ids,
            channel_names=channel_names,
            overwrite=overwrite,
            ome_name=ome_name,
        )
        # no output file name means we are skipping
        if tmp_output_file_name is None:
            return Path(output_file_name) if output_file_name else None

        assert channel_names is not None, "Channel names must be defined"
        assert channel_ids is not None, "Channel ids must be defined"

        # write OME-XML to the ImageDescription tag of the first page
        description = self.omexml

        reader = self.reader
        rgb_image = np.zeros((self.y_size, self.x_size, 3), dtype=np.uint8 if as_uint8 else reader.dtype)
        with TiffWriter(tmp_output_file_name, bigtiff=True) as tif, MeasureTimer() as main_timer:
            for i, (_channel_name, _channel_index, image) in enumerate(
                self._processed_image_yield(channel_ids, channel_names, as_uint8)
            ):
                rgb_image[:, :, i] = image

            # write channel data
            msg = "Writing RGB image"
            past_msg = msg.replace("Writing", "Wrote")
            # write channel data
            logger.trace(f"{msg} - {image.shape}...")  # type: ignore[attr-defined]

            with MeasureTimer() as write_timer:
                tif.write(rgb_image, subifds=self.subifds, description=description, **options)
                logger.trace(f"{past_msg} pyramid index 0 in {write_timer()}")

                if write_pyramid:
                    logger.info("Writing pyramid...")
                    for pyramid_index, pyr_image in self._convert_image_to_pyramid(rgb_image):
                        logger.trace(f"{msg} pyramid index {pyramid_index} - {pyr_image.shape}...")
                        tif.write(pyr_image, **options, subfiletype=1)
                        logger.trace(f"{past_msg} pyramid index {pyramid_index} in {write_timer(since_last=True)}")
                        del pyr_image
                del rgb_image
        logger.trace(f"Exported OME-TIFF in {main_timer()}")
        # rename tmp file to output file
        retry(lambda: tmp_output_file_name.rename(output_file_name), PermissionError)()  # type: ignore[arg-type]
        logger.trace(f"Renamed tmp file to output file ({output_file_name})")
        return Path(output_file_name)

    def _write_image_multichannel(
        self,
        name: str,
        output_dir: Path | str = "",
        write_pyramid: bool = True,
        tile_size: int = 512,
        compression: str | None = "default",
        as_uint8: bool | None = None,
        channel_ids: list[int | tuple[int, ...]] | None = None,
        channel_names: list[str] | None = None,
        overwrite: bool = False,
        ome_name: str | None = None,
    ) -> Path | None:
        (
            tmp_output_file_name,
            output_file_name,
            options,
            channel_ids,
            channel_names,
            as_uint8,
            write_pyramid,
        ) = self._prepare_tiff(
            name=name,
            output_dir=output_dir,
            write_pyramid=write_pyramid,
            tile_size=tile_size,
            compression=compression,
            as_uint8=as_uint8,
            channel_ids=channel_ids,
            channel_names=channel_names,
            overwrite=overwrite,
            ome_name=ome_name,
        )
        # no output file name means we are skipping
        if tmp_output_file_name is None:
            return Path(output_file_name) if output_file_name else None

        assert channel_names is not None, "Channel names must be defined"
        assert channel_ids is not None, "Channel ids must be defined"

        # write OME-XML to the ImageDescription tag of the first page
        description = self.omexml

        with TiffWriter(tmp_output_file_name, bigtiff=True) as tif, MeasureTimer() as main_timer:
            for channel_name, channel_index, image in self._processed_image_yield(channel_ids, channel_names, as_uint8):
                msg = f"Writing image for channel={channel_name} ({channel_index})"
                past_msg = msg.replace("Writing", "Wrote")
                # write channel data
                logger.trace(f"{msg} - {image.shape}...")
                with MeasureTimer() as write_timer:
                    tif.write(image, subifds=self.subifds, description=description, **options)
                    logger.trace(f"{past_msg} pyramid index 0 in {write_timer()}")
                    if write_pyramid:
                        for pyramid_index, pyr_image in self._convert_image_to_pyramid(image):
                            logger.trace(f"{msg} pyramid index {pyramid_index} - {pyr_image.shape}...")
                            tif.write(pyr_image, **options, subfiletype=1)
                            logger.trace(f"{past_msg} pyramid index {pyramid_index} in {write_timer(since_last=True)}")
                            del pyr_image
                    del image
        logger.trace(f"Exported OME-TIFF in {main_timer()}")
        # rename tmp file to output file
        retry(lambda: tmp_output_file_name.rename(output_file_name), PermissionError)()  # type: ignore[arg-type]
        logger.trace(f"Renamed tmp file to output file ({output_file_name})")
        return Path(output_file_name)

    def write_image_by_plane(
        self,
        name: str,
        output_dir: Path | str = "",
        write_pyramid: bool = True,
        tile_size: int = 512,
        compression: str | None = "default",
        as_uint8: bool | None = None,
        channel_ids: list[int | tuple[int, ...]] | None = None,
        channel_names: list[str] | None = None,
        overwrite: bool = False,
        ome_name: str | None = None,
    ) -> Path | None:
        """Write OME-TIFF image plane-by-plane to disk.

        WsiReg compatible RegImages all have methods to read an image channel-by-channel, thus each channel is read,
        transformed, and written to reduce memory during file writing.

        RGB images may run large memory footprints as they are interleaved before write, for RGB images,
        using the `OmeTiledTiffWriter` is recommended.

        Parameters
        ----------
        name: str
            Name to be written WITHOUT extension
            for example if image_name = "cool_image" the file
            would be "cool_image.ome.tiff"
        output_dir: Path or str
            Directory where the image will be saved
        write_pyramid: bool
            Whether to write the OME-TIFF with sub-resolutions or not
        tile_size: int
            What size to write OME-TIFF tiles to disk
        compression: str
            tifffile string to pass to compression argument, defaults to "deflate" for minisblack
            and "jpeg" for RGB type images
        as_uint8: bool
            Whether to save the image in the 0-255 intensity range, substantially reducing file size at the cost of some
            intensity information.
        channel_ids: list of int
            Channel indices to write to OME-TIFF, if None, all channels are written
        channel_names: list of str
            Channel names.
        overwrite: bool
            Whether to overwrite the file if it already exists
        ome_name: str
            Name to be written to the OME-XML metadata, if None, the name is the same as the image name

        Returns
        -------
        output_file_name: str
            File path to the written OME-TIFF
        """
        # make sure user did not provide filename with OME-TIFF
        reader = self.reader
        if reader.is_rgb:
            return self._write_image_rgb(
                name,
                output_dir,
                write_pyramid=write_pyramid,
                tile_size=tile_size,
                compression=compression,
                as_uint8=as_uint8,
                channel_ids=channel_ids,
                channel_names=channel_names,
                overwrite=overwrite,
                ome_name=ome_name,
            )
        else:
            return self._write_image_multichannel(
                name,
                output_dir,
                write_pyramid=write_pyramid,
                tile_size=tile_size,
                compression=compression,
                as_uint8=as_uint8,
                channel_ids=channel_ids,
                channel_names=channel_names,
                overwrite=overwrite,
                ome_name=ome_name,
            )


class OmeTiffWrapper:
    """OME-TIFF wrapper."""

    reader: BaseReader
    writer: OmeTiffWriter
    tiff: TiffWriter
    path: Path

    as_uint8: bool
    description: str
    options: dict
    write_pyramid: bool

    @property
    def is_rgb(self) -> bool:
        """Check if image is RGB."""
        return self.reader.is_rgb

    @property
    def resolution(self) -> float:
        """Resolution."""
        return self.reader.resolution

    @contextmanager
    def write(
        self,
        channel_names: list[str],
        resolution: float,
        shape: tuple[int, ...],
        dtype: np.dtype,
        name: str,
        output_dir: Path | str = "",
        write_pyramid: bool = True,
        tile_size: int = 512,
        compression: str | None = "default",
        as_uint8: bool | None = None,
        overwrite: bool = False,
        transformer: Transformer | None = None,
    ) -> ty.Generator[OmeTiffWrapper, None, None]:
        """Write."""
        from image2image_io.readers._base_reader import DummyReader

        self.reader = DummyReader(
            channel_names=channel_names,
            resolution=resolution,
            shape=shape,
            dtype=dtype,
        )
        self.writer = OmeTiffWriter(reader=self.reader, transformer=transformer)
        (
            tmp_output_file_name,
            output_file_name,
            self.options,
            channel_ids,
            channel_names,
            self.as_uint8,
            self.write_pyramid,
        ) = self.writer._prepare_tiff(
            name=name,
            output_dir=output_dir,
            write_pyramid=write_pyramid,
            tile_size=tile_size,
            compression=compression,
            as_uint8=as_uint8,
            channel_names=channel_names,
            overwrite=overwrite,
        )

        # no output file name means we are skipping
        if tmp_output_file_name is None:
            raise ValueError("No output file name")
        assert channel_names is not None, "Channel names must be defined"
        assert channel_ids is not None, "Channel ids must be defined"

        # write OME-XML to the ImageDescription tag of the first page
        self.description = self.writer.omexml

        # write channel data
        with TiffWriter(tmp_output_file_name, bigtiff=True) as self.tiff, MeasureTimer() as timer:
            yield self

        logger.trace(f"Exported OME-TIFF in {timer()}")
        # rename tmp file to output file
        retry(lambda: tmp_output_file_name.rename(output_file_name), PermissionError)()  # type: ignore[arg-type]
        logger.trace(f"Renamed tmp file to output file ({output_file_name})")
        self.path = output_file_name

    def add_channel(self, channel_index: int | list[int], channel_name: str | list[str], array: np.ndarray) -> None:
        """Add channel."""
        assert self.reader is not None, "Reader must be defined"
        assert self.writer is not None, "Writer must be defined"
        assert self.tiff is not None, "Tiff writer must be defined"
        if self.is_rgb:
            self._add_rgb_channel(channel_index, channel_name, array)
        else:
            self._add_multichannel_channel(channel_index, channel_name, array)

    def _add_rgb_channel(self, channel_index: list[int], channel_name: list[str], array: np.ndarray) -> None:
        array = np.dstack(
            [
                self.writer._process_image(array[:, :, index], resolution=self.resolution, as_uint8=self.as_uint8)
                for index in channel_index
            ]
        )
        msg = f"Writing image for channel={channel_name} ({channel_index})"
        past_msg = msg.replace("Writing", "Wrote")
        # write channel data
        logger.trace(f"{msg} - {array.shape}...")
        with MeasureTimer() as write_timer:
            self.tiff.write(array, subifds=self.writer.subifds, description=self.description, **self.options)
            logger.trace(f"{past_msg} pyramid index 0 in {write_timer()}")

            if self.write_pyramid:
                for pyramid_index, pyr_image in self.writer._convert_image_to_pyramid(array):
                    logger.trace(f"{msg} pyramid index {pyramid_index} - {pyr_image.shape}...")
                    self.tiff.write(pyr_image, **self.options, subfiletype=1)
                    logger.trace(f"{past_msg} pyramid index {pyramid_index} in {write_timer(since_last=True)}")
                    del pyr_image
            del array

    def _add_multichannel_channel(self, channel_index: int, channel_name: str, array: np.ndarray) -> None:
        msg = f"Writing image for channel={channel_name} ({channel_index})"
        past_msg = msg.replace("Writing", "Wrote")
        # write channel data
        logger.trace(f"{msg} - {array.shape}...")
        with MeasureTimer() as write_timer:
            array = self.writer._process_image(array, resolution=self.resolution, as_uint8=self.as_uint8)
            self.tiff.write(array, subifds=self.writer.subifds, description=self.description, **self.options)
            logger.trace(f"{past_msg} pyramid index 0 in {write_timer()}")

            if self.write_pyramid:
                for pyramid_index, pyr_image in self.writer._convert_image_to_pyramid(array):
                    logger.trace(f"{msg} pyramid index {pyramid_index} - {pyr_image.shape}...")
                    self.tiff.write(pyr_image, **self.options, subfiletype=1)
                    logger.trace(f"{past_msg} pyramid index {pyramid_index} in {write_timer(since_last=True)}")
                    del pyr_image
            del array



"""Create thumbnail for image(s)."""

from __future__ import annotations

import click
from koyo.click import cli_parse_paths_sort
from loguru import logger
from tqdm import tqdm


@click.option(
    "-f/-F",
    "--first_only/--no_first_only",
    help="Export only the first channel from the stack.",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "-t/-T",
    "--with_title/--no_with_title",
    help="Export image with title.",
    is_flag=True,
    default=True,
    show_default=True,
)
@click.option(
    "-o",
    "--output_dir",
    help="Path to directory where OME-TIFF files will be saved.",
    default=".",  # cwd
    type=click.Path(exists=False, resolve_path=True, file_okay=False, dir_okay=True),
    show_default=True,
    required=True,
)
@click.option(
    "-i",
    "--input",
    "input_",
    help="Path to supported file(s).",
    type=click.UNPROCESSED,
    show_default=True,
    required=True,
    multiple=True,
    callback=cli_parse_paths_sort,
)
@click.command()
def thumbnail(input_: str, output_dir: str, with_title: bool, first_only: bool) -> None:
    """Create a thumbnail for image(s)."""
    from image2image_io.utils.utilities import write_thumbnail

    for path in tqdm(input_, desc="Creating thumbnails..."):
        write_thumbnail(path, output_dir, with_title=with_title, first_only=first_only)
        logger.info(f"Created thumbnail for '{path}'.")



"""TIFF file utilities.

Taken from:
https://github.com/NHPatterson/napari-imsmicrolink/blob/master/src/napari_imsmicrolink/utils/tifffile_meta.py
"""

from __future__ import annotations

from ome_types.model import OME
from tifffile import TiffFile


def qptiff_channel_names(tif: TiffFile, series_index: int = 0) -> list[str]:
    """Retrieve filenames from qptiff."""
    from image2image_io.utils.utilities import xmlstr_to_dict

    channel_names: list[str] = []
    pages = tif.series[series_index].pages
    if not pages:
        return channel_names
    for index, page in enumerate(pages):
        xml = page.tags["ImageDescription"].value
        xml_dict = xmlstr_to_dict(xml)
        if "PerkinElmer-QPI-ImageDescription" in xml_dict:
            xml_dict = xml_dict["PerkinElmer-QPI-ImageDescription"]
            channel_name = []
            if "Biomarker" in xml_dict:
                channel_name.append(xml_dict["Biomarker"])
            if "Name" in xml_dict:
                channel_name.append(xml_dict["Name"])
            channel_name = " - ".join(channel_name)  # type: ignore[assignment]
            channel_name += f" (C{index})"
            channel_names.append(channel_name)  # type: ignore[arg-type]
    return channel_names


def tifftag_xy_pixel_sizes(rdr: TiffFile, series_idx: int, level_idx: int) -> tuple[float, ...]:
    """Resolution data is stored in the TIFF tags in pixels per cm, this is converted to microns per pixel."""
    # pages are accessed because they contain the tiff tag
    # subset by series -> level -> first page contains all tags
    current_page = rdr.series[series_idx].levels[level_idx].pages[0]

    x_res = current_page.tags["XResolution"].value
    y_res = current_page.tags["XResolution"].value

    res_unit = current_page.tags["ResolutionUnit"].value

    # convert units to micron
    # res_unit == 1: undefined (px)
    # res_unit == 2 pixels per inch
    # res unit == 3 pixels per cm
    # in all cases we convert to um
    # https://www.awaresystems.be/imaging/tiff/tifftags/resolutionunit.html
    if res_unit.value == 1:
        res_to_um = 1
    if res_unit.value == 2:
        res_to_um = 25400
    elif res_unit.value == 3:
        res_to_um = 10000

    # conversion of pixels / um to um / pixel
    x_res_um = (1 / (x_res[0] / x_res[1])) * res_to_um
    y_res_um = (1 / (y_res[0] / y_res[1])) * res_to_um
    return x_res_um, y_res_um


def svs_xy_pixel_sizes(rdr: TiffFile, series_idx: int, level_idx: int) -> tuple[float, ...]:
    """Get resolution data stored in ImageDescription of SVS."""
    # pages are accessed because they contain the tiff tag
    # subset by series -> level -> first page contains all tags
    current_page = rdr.series[series_idx].levels[level_idx].pages[0]
    id_str = current_page.tags["ImageDescription"].value
    svs_info = id_str.split("|")
    mpp_val = None
    for i in svs_info:
        if "MPP" in i:
            mpp_val = float(i.split("=")[1])

    if mpp_val:
        return mpp_val, mpp_val
    return 1.0, 1.0


def ometiff_xy_pixel_sizes(ome_metadata: OME, series_idx: int):
    """Get resolution data stored in OME metadata."""
    from pint import UnitRegistry

    ps_x = ome_metadata.images[series_idx].pixels.physical_size_x
    ps_y = ome_metadata.images[series_idx].pixels.physical_size_y
    ps_unit = ome_metadata.images[series_idx].pixels.physical_size_x_unit

    if ps_unit.name.lower() != "micrometer":
        ureg = UnitRegistry()
        cur_size = ps_x * ureg(ps_unit.name.lower())
        out_um = cur_size.to("micrometer")
        ps_x_out = out_um.magnitude
        return (ps_x_out, ps_x_out)
    else:
        return (ps_x, ps_y)


def ometiff_ch_names(ome_metadata: OME, series_idx: int):
    """Get channel names stored in OME metadata."""
    ch_meta = ome_metadata.images[series_idx].pixels.channels
    cnames = []
    for idx, ch in enumerate(ch_meta):
        if ch.name:
            cnames.append(ch.name)
        else:
            cnames.append(f"C{str(idx + 1).zfill(2)}")

    return cnames



"""ITK utilities."""

from copy import deepcopy

import numpy as np

AFFINE_TRANSFORM = {
    "Transform": ["AffineTransform"],
    "NumberOfParameters": ["6"],
    "TransformParameters": ["1", "0", "0", "1", "0", "0"],
    "InitialTransformParametersFileName": ["NoInitialTransform"],
    "HowToCombineTransforms": ["Compose"],
    "FixedImageDimension": ["2"],
    "MovingImageDimension": ["2"],
    "FixedInternalImagePixelType": ["float"],
    "MovingInternalImagePixelType": ["float"],
    "Size": ["0", "0"],
    "Index": ["0", "0"],
    "Spacing": ["0", "0"],
    "Origin": ["0.0000", "0.0000"],
    "Direction": [
        "1.0000000000",
        "0.0000000000",
        "0.0000000000",
        "1.0000000000",
    ],
    "UseDirectionCosines": ["true"],
    "CenterOfRotationPoint": ["0", "0"],
    "ResampleInterpolator": ["FinalNearestNeighborInterpolator"],
    "Resampler": ["DefaultResampler"],
    "DefaultPixelValue": ["0.000000"],
    "ResultImageFormat": ["mha"],
    "ResultImagePixelType": ["float"],
    "CompressResultImage": ["true"],
}


def compute_affine_bound(
    shape: tuple[int, int], affine: np.ndarray, spacing: float = 1
) -> tuple[tuple[float, float], tuple[float, float]]:
    """Compute affine bounds."""
    w, h = shape
    # Top-left, Top-right, Bottom-left, Bottom-right
    corners = np.array(
        [
            [0, 0, 1],  # Adding 1 for homogeneous coordinates (x, y, 1)
            [w, 0, 1],
            [0, h, 1],
            [w, h, 1],
        ]
    )
    # Apply affine transformation to corners
    transformed_corners = np.dot(corners, affine.T)  # Transpose matrix to match shapes

    # Extracting x and y coordinates
    x_coords, y_coords = transformed_corners[:, 0], transformed_corners[:, 1]

    # Calculate bounds
    min_x, max_x = np.min(x_coords), np.max(x_coords)
    min_y, max_y = np.min(y_coords), np.max(y_coords)

    # Calculate new  width and height
    new_width = int(np.ceil(max_x - min_x))
    new_height = int(np.ceil(max_y - min_y))

    affine_ = affine[:2, :]
    new_width += abs(affine_[1, 2] / spacing)
    new_height += abs(affine_[0, 2] / spacing)
    return (new_width, new_height), (new_width / 2, new_height / 2)


def affine_to_itk_affine(
    affine: np.ndarray,
    image_shape: tuple[int, int],
    spacing: float = 1.0,
    inverse: bool = False,
) -> dict:
    """Convert affine matrix (yx, um) to ITK affine matrix.

    The assumption is that the affine matrix is provided in numpy ordering (e.g. from napari) and values are in um.
    """
    # TODO change the origin so that we don't have to make the image bigger

    assert affine.shape == (3, 3), "affine matrix must be 3x3"
    if inverse:
        affine = np.linalg.inv(affine)

    tform = deepcopy(AFFINE_TRANSFORM)
    tform["Spacing"] = [str(spacing), str(spacing)]
    # compute new image shape
    (bound_w, bound_h), (origin_x, origin_y) = compute_affine_bound(image_shape, affine, spacing)  # width, height

    # calculate rotation center point
    # center_of_rot = calculate_center_of_rotation(affine, image_shape, (spacing, spacing))
    # center_of_rot = image.TransformContinuousIndexToPhysicalPoint(
    #     ((bound_w - 1) / 2, (bound_h - 1) / 2),
    # )  # type: ignore[no-untyped-call]
    # center_of_rot = ((bound_w - 1) / 2, (bound_h - 1) / 2)
    # tform["CenterOfRotationPoint"] = [str(center_of_rot[0]), str(center_of_rot[1])]

    # adjust for pixel spacing
    tform["Size"] = [str(int(np.ceil(bound_w))), str(int(np.ceil(bound_h)))]
    # tform["Origin"] = [str(origin_x), str(origin_y)]

    # extract affine parameters
    affine_ = affine[:2, :]
    # a, b, c, d, tx, ty
    tform["TransformParameters"] = [
        affine_[1, 1],
        affine_[1, 0],
        affine_[0, 1],
        affine_[0, 0],
        affine_[1, 2],
        affine_[0, 2],
    ]
    return tform



"""Tiff file utilities."""

from __future__ import annotations

import typing as ty
from pathlib import Path

import dask.array as da
import numpy as np
import pandas as pd
import SimpleITK as sitk
import zarr
from koyo.timer import MeasureTimer
from loguru import logger
from tifffile import TiffFile, imread, xml2dict

from image2image_io.utils.utilities import guess_rgb

logger = logger.bind(src="Tiff")


def tifffile_to_dask(im_fp: str | Path, largest_series: int) -> list[da.Array]:
    """Convert a tifffile to a dask array."""
    with MeasureTimer() as timer:
        imdata = zarr.open(imread(im_fp, aszarr=True, series=largest_series))
    logger.trace(f"Loaded OME-TIFF data in {timer()}")
    with MeasureTimer() as timer:
        if isinstance(imdata, zarr.hierarchy.Group):
            imdata = [da.from_zarr(imdata[z]) for z in imdata.array_keys()]
        else:
            imdata = [da.from_zarr(imdata)]
    logger.trace(f"Converted OME-TIFF data to dask in {timer()}")
    return imdata


def calc_pyramid_levels(xy_final_shape: np.ndarray, tile_size: int) -> list[tuple[int, int]]:
    """Calculate the number of pyramids for a given image dimension, and tile size.

    Stops when further downsampling would be smaller than tile_size.

    Parameters
    ----------
    xy_final_shape:np.ndarray
        final shape in xy order
    tile_size: int
        size of the tiles in the pyramidal layers

    Returns
    -------
    res_shapes:list
        list of tuples of the shapes of the downsampled images

    """
    res_shape = xy_final_shape[::-1]
    res_shapes = [(int(res_shape[0]), int(res_shape[1]))]

    while all(res_shape > tile_size):
        res_shape = res_shape // 2
        res_shapes.append((int(res_shape[0]), int(res_shape[1])))
    return res_shapes[:-1]


def get_pyramid_info(
    y_size: int, x_size: int, n_ch: int, tile_size: int
) -> tuple[list[tuple[int, int]], list[tuple[int, int, int, int, int]]]:
    """
    Get pyramidal info for OME-tiff output.

    Parameters
    ----------
    y_size: int
        y dimension of base layer
    x_size:int
        x dimension of base layer
    n_ch:int
        number of channels in the image
    tile_size:int
        tile size of the image

    Returns
    -------
    pyr_levels
        pyramidal levels
    pyr_shapes:
        OME-zarr pyramid shapes for all levels

    """
    yx_size = np.asarray([y_size, x_size], dtype=np.int32)
    pyr_levels = calc_pyramid_levels(yx_size, tile_size)
    pyr_shapes = [(1, n_ch, 1, int(pl[0]), int(pl[1])) for pl in pyr_levels]
    return pyr_levels, pyr_shapes


def centered_transform(
    image_size,
    image_spacing,
    rotation_angle,
) -> np.ndarray:
    """Centered rotation transform."""
    angle = np.deg2rad(rotation_angle)

    sina = np.sin(angle)
    cosa = np.cos(angle)

    # build rot mat
    rot_mat = np.eye(3)
    rot_mat[0, 0] = cosa
    rot_mat[1, 1] = cosa
    rot_mat[1, 0] = sina
    rot_mat[0, 1] = -sina

    # recenter transform
    center_point = np.multiply(image_size, image_spacing) / 2
    center_point = np.append(center_point, 0)
    translation = center_point - np.dot(rot_mat, center_point)
    rot_mat[:2, 2] = translation[:2]

    return rot_mat


def get_largest_pyramid_series_tiff(image_filepath):
    """Determine the largest series for .scn files by examining metadata.

    For other multi-series files, find the one with the most pixels.

    Parameters
    ----------
    image_filepath: str
        path to the image file

    Returns
    -------
    largest_series:int
        index of the largest series in the image data
    """
    fp_ext = Path(image_filepath).suffix.lower()
    tf_im = TiffFile(image_filepath)
    if fp_ext == ".scn":
        scn_meta = xml2dict(tf_im.scn_metadata)
        image_meta = scn_meta.get("scn").get("collection").get("image")
        largest_series = np.argmax(
            [im.get("scanSettings").get("objectiveSettings").get("objective") for im in image_meta]
        )
    else:
        largest_series = np.argmax([np.prod(np.asarray(series.shape), dtype=np.int64) for series in tf_im.series])
    tf_im.close()
    tf_im.filehandle.close()
    del tf_im
    return largest_series


def tifffile_zarr_backend(image_filepath, largest_series, preprocessing, force_rgb=None):
    """
    Read image with tifffile and use zarr to read data into memory.

    Parameters
    ----------
    image_filepath: str
        path to the image file
    largest_series: int
        index of the largest series in the image
    preprocessing:
        whether to do some read-time pre-processing
        - greyscale conversion (at the tile level)
        - read individual or range of channels (at the tile level)
    force_rgb: bool
        whether to force the image to be read as rgb

    Returns
    -------
    image: sitk.Image
        image ready for another registration pre-processing

    """
    zarr_series = imread(image_filepath, aszarr=True, series=largest_series)
    zarr_store = zarr.open(zarr_series)
    zarr_im = zarr_get_base_pyr_layer(zarr_store)
    return read_preprocess_array(zarr_im, preprocessing=preprocessing, force_rgb=force_rgb)


def tifffile_dask_backend(image_filepath, largest_series, preprocessing, force_rgb=None):
    """
    Read image with tifffile and use dask to read data into memory.

    Parameters
    ----------
    image_filepath: str
        path to the image file
    largest_series: int
        index of the largest series in the image
    preprocessing:
        whether to do some read-time pre-processing
        - greyscale conversion (at the tile level)
        - read individual or range of channels (at the tile level)
    force_rgb: bool
        whether to force the image to be read as rgb

    Returns
    -------
    image: sitk.Image
        image ready for other registration pre-processing

    """
    zarr_series = imread(image_filepath, aszarr=True, series=largest_series)
    zarr_store = zarr.open(zarr_series)
    dask_im = da.squeeze(da.from_zarr(zarr_get_base_pyr_layer(zarr_store)))
    return read_preprocess_array(dask_im, preprocessing=preprocessing, force_rgb=force_rgb)


def zarr_get_base_pyr_layer(zarr_store):
    """
    Find the base pyramid layer of a zarr store.

    Parameters
    ----------
    zarr_store
        zarr store

    Returns
    -------
    zarr_im: zarr.core.Array
        zarr array of base layer
    """
    if isinstance(zarr_store, zarr.hierarchy.Group):
        zarr_im = zarr_store[str(0)]
    elif isinstance(zarr_store, zarr.core.Array):
        zarr_im = zarr_store
    else:
        raise ValueError("Could not find base pyramid layer.")
    return zarr_im


def ensure_dask_array(image):
    """Ensure array is a dask array."""
    if isinstance(image, da.core.Array):
        return image

    if isinstance(image, zarr.Array):
        return da.from_zarr(image)

    # handles np.ndarray _and_ other array like objects.
    return da.from_array(image)


def read_preprocess_array(array, preprocessing, force_rgb=None):
    """Read np.array, zarr.Array, or dask.array image into memory with preprocessing for registration."""
    is_interleaved = guess_rgb(array.shape)
    is_rgb = is_interleaved if not force_rgb else force_rgb

    if is_rgb:
        if preprocessing:
            image_out = np.asarray(grayscale(array, is_interleaved=is_interleaved))
            image_out = sitk.GetImageFromArray(image_out)
        else:
            image_out = np.asarray(array)
            if not is_interleaved:
                image_out = np.rollaxis(image_out, 0, 3)
            image_out = sitk.GetImageFromArray(image_out, isVector=True)

    elif len(array.shape) == 2:
        image_out = sitk.GetImageFromArray(np.asarray(array))

    else:
        if preprocessing:
            if preprocessing.ch_indices and len(array.shape) > 2:
                chs = list(preprocessing.ch_indices)
                array = array[chs, :, :]

        image_out = sitk.GetImageFromArray(np.squeeze(np.asarray(array)))

    return image_out


def get_tifffile_info(image_filepath):
    """Get tiff file information."""
    largest_series = get_largest_pyramid_series_tiff(image_filepath)
    zarr_im = zarr.open(imread(image_filepath, aszarr=True, series=largest_series))
    zarr_im = zarr_get_base_pyr_layer(zarr_im)
    image_shape = image_shape_ = np.squeeze(zarr_im.shape)
    if len(image_shape_) == 2:
        image_shape_ = np.concatenate([[1], image_shape_])
    return image_shape, image_shape_, zarr_im.dtype, largest_series


def grayscale(rgb_image: np.ndarray | da.Array, is_interleaved: bool = False) -> np.ndarray:
    """
    Convert RGB image data to greyscale.

    Parameters
    ----------
    rgb_image: np.ndarray
        image data
    is_interleaved: bool
        whether the image is interleaved

    Returns
    -------
    image:np.ndarray
        returns 8-bit greyscale image for 24-bit RGB image
    """
    if is_interleaved:
        result = (
            (rgb_image[..., 0] * 0.2125).astype(np.uint8)
            + (rgb_image[..., 1] * 0.7154).astype(np.uint8)
            + (rgb_image[..., 2] * 0.0721).astype(np.uint8)
        )
    else:
        result = (
            (rgb_image[0, ...] * 0.2125).astype(np.uint8)
            + (rgb_image[1, ...] * 0.7154).astype(np.uint8)
            + (rgb_image[2, ...] * 0.0721).astype(np.uint8)
        )
    return result


def tf_zarr_read_single_ch(image_filepath, channel_idx, is_rgb, is_rgb_interleaved=True):
    """
    Reads a single channel using zarr or dask in combination with tifffile.

    Parameters
    ----------
    image_filepath:str
        file path to image
    channel_idx:int
        index of the channel to be read
    is_rgb:bool
        whether image is rgb interleaved
    is_rgb_interleaved:bool
        whether image is rgb interleaved

    Returns
    -------
    im:np.ndarray
        image as a np.ndarray
    """
    largest_series = get_largest_pyramid_series_tiff(image_filepath)
    zarr_im = zarr.open(imread(image_filepath, aszarr=True, series=largest_series))
    zarr_im = zarr_get_base_pyr_layer(zarr_im)
    try:
        im = da.squeeze(da.from_zarr(zarr_im))
        if is_rgb and is_rgb_interleaved:
            im = im[:, :, channel_idx].compute()
        elif len(im.shape) > 2:
            im = im[channel_idx, :, :].compute()
        else:
            im = im.compute()

    except ValueError:
        im = zarr_im
        if is_rgb is True and is_rgb_interleaved:
            im = im[:, :, channel_idx]
        elif len(im.shape) > 2:
            im = im[channel_idx, :, :].compute()
        else:
            im = im.compute()
    return im


def yield_tiles(z, tile_size, is_rgb):
    """Return tiles."""
    for y in range(0, z.shape[0], tile_size):
        for x in range(0, z.shape[1], tile_size):
            if is_rgb:
                yield z[y : y + tile_size, x : x + tile_size, :].compute()


def compute_sub_res(zarray, ds_factor, tile_size, is_rgb, im_dtype) -> da.Array:
    """Compute sub-resolution image."""
    if is_rgb:
        resampling_axis = {0: 2**ds_factor, 1: 2**ds_factor, 2: 1}
        tiling = (tile_size, tile_size, 3)
    else:
        resampling_axis = {0: 1, 1: 2**ds_factor, 2: 2**ds_factor}
        tiling = (1, tile_size, tile_size)

    resampled_zarray_subres = da.coarsen(np.mean, zarray, resampling_axis, trim_excess=True)
    resampled_zarray_subres = resampled_zarray_subres.astype(im_dtype)
    resampled_zarray_subres = resampled_zarray_subres.rechunk(tiling)
    return resampled_zarray_subres


def prepare_ome_xml_str(
    y_size: int, x_size: int, n_ch: int, im_dtype: np.dtype[ty.Any], is_rgb: bool, **ome_metadata: ty.Any
) -> str:
    """Prepare OME-XML metadata."""
    from tifffile import OmeXml

    omexml = OmeXml()
    if is_rgb:
        stored_shape = (1, 1, 1, y_size, x_size, n_ch)
        im_shape = (y_size, x_size, n_ch)
    else:
        stored_shape = (n_ch, 1, 1, y_size, x_size, 1)
        im_shape = (n_ch, y_size, x_size)

    omexml.addimage(
        dtype=im_dtype,
        shape=im_shape,
        # specify how the image is stored in the TIFF file
        storedshape=stored_shape,
        **ome_metadata,
    )

    return omexml.tostring().encode("utf8")


def check_df_columns(
    df: pd.DataFrame,
    required: list[str],
    either: list[tuple[str, ...]] | None = None,
    either_dtype: tuple[np.dtype, ...] | None = None,
) -> bool:
    """Check if a DataFrame has the required columns."""
    if not all(col in df.columns for col in required):
        return False
    if either is not None:
        if either_dtype is not None:
            return all(any((col in df.columns) and (df[col].dtype in either_dtype) for col in cols) for cols in either)
        return all(any(col in df.columns for col in cols) for cols in either)
    return True


def get_column_name(df: pd.DataFrame, options: list[str]) -> str:
    """Get columns from a DataFrame."""
    for key in options:
        if key in df.columns:
            return key
    raise ValueError(f"None of the options {options} are in the DataFrame. Available columns are {df.columns}")



"""Utilities."""

from __future__ import annotations

import typing as ty
from contextlib import suppress
from pathlib import Path
from tempfile import tempdir

from koyo.typing import PathLike

TEST_DATA_DIR = Path(__file__).parent.parent.parent.parent / "tests" / "_test_data"


# noinspection PyMissingOrEmptyDocstring
def get_test_files(pattern: str = "*.json") -> list[Path]:
    """Get config files."""
    return list(TEST_DATA_DIR.glob(pattern))


def get_test_file(filename: str) -> Path:
    """Get config file."""
    path = TEST_DATA_DIR / filename
    assert path.exists(), f"File {path} does not exist."
    return path


def make_test_polygon(output_dir: Path) -> None:
    """Create test data."""
    import numpy as np
    import tifffile
    from skimage.draw import polygon2mask

    im = polygon2mask(
        (2048, 2048),
        polygon=np.array(
            (
                (300, 300),
                (980, 320),
                (380, 730),
                (220, 590),
                (300, 300),
            )
        ),
    )
    im = im.astype(np.uint8)
    im[im == 1] = 255
    tifffile.imwrite(output_dir / "moving.tiff", im, compression="deflate")

    im = polygon2mask(
        (2048, 2048),
        polygon=np.array(
            (
                (300, 300),
                (980, 320),
                (380, 730),
                (220, 590),
                (300, 300),
            )
        )
        + 900,
    )
    im = im.astype(np.uint8)
    im[im == 1] = 255
    tifffile.imwrite(output_dir / "fixed.tiff", im, compression="deflate")


class CleanupMixin:
    """Mixin class."""

    _temp_temp = None
    _base_temp = None

    @property
    def temp_dir(self) -> Path:
        """Temporary directory."""
        if self._temp_temp is None:
            self._temp_temp = Path(tempdir) / "temp"
            self._temp_temp.mkdir()
        return self._temp_temp

    def setup_method(self, method: ty.Any) -> None:
        """Setup class."""
        if self._base_temp is None:
            self._base_temp = Path(tempdir) / "autoims"
            self._base_temp.mkdir(exist_ok=True)

    def teardown_method(self, method: ty.Any) -> None:
        """Teardown class."""
        if self._base_temp:
            cleanup_directory(self._base_temp)
        if self._temp_temp:
            cleanup_directory(self._temp_temp)


def cleanup_directory(path: PathLike) -> None:
    """Clean directory."""
    import glob
    import os
    import shutil

    try:
        shutil.rmtree(path)
    except Exception:
        for f in glob.glob(os.path.join(path, "*"), recursive=True):
            if os.path.isdir(f):
                cleanup_directory(f)
            else:
                with suppress(PermissionError):
                    os.remove(f)



"""Create mask(s) for image(s) based on shapes."""

from __future__ import annotations

from datetime import datetime
from pathlib import Path

import cv2
import numpy as np
from koyo.timer import MeasureTimer
from koyo.typing import PathLike
from loguru import logger
from shapely import Polygon
from tqdm import tqdm

from image2image_io.enums import TIME_FORMAT, MaskOutputFmt

logger = logger.bind(src="Mask")


def mask_to_polygon(mask: np.ndarray, epsilon: float = 1) -> np.ndarray:
    """Convert mask to polygon."""
    mask = np.asarray(mask, dtype=np.uint8)
    # ensure mask is between 0-255
    if mask.max() == 1:
        mask = mask * 255

    # Find contours
    contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    # Approximate contours to polygons and get the coordinates
    polygons = [cv2.approxPolyDP(contour, epsilon=epsilon, closed=True) for contour in contours]
    polygons = np.asarray(polygons)
    return polygons


def shapes_to_polygons(
    shapes: list[np.ndarray], with_index: bool = False, inv_pixel_size: float = 1.0, silent: bool = True
) -> list[Polygon] | list[tuple[Polygon, int]]:
    """Convert shapes to polygons."""
    polygons = []
    for index, shape in enumerate(
        tqdm(shapes, desc="Drawing shapes", leave=False, miniters=1000, disable=silent), start=1
    ):
        if isinstance(shape, dict):
            shape = shape["array"]
        yx = shape * inv_pixel_size
        if with_index:
            polygons.append((Polygon(yx), index))
        else:
            polygons.append(Polygon(yx))
    return polygons


def polygons_to_mask(polygons: list[Polygon], output_shape: tuple[int, int]) -> np.ndarray:
    """Convert polygons to mask."""
    from rasterio.features import rasterize

    mask: np.ndarray = rasterize(polygons, out_shape=output_shape)
    return mask


def transform_mask(mask: np.ndarray, transform: np.ndarray, output_shape: tuple[int, int]) -> np.ndarray:
    """Transform mask."""
    from scipy.ndimage import affine_transform
    from skimage.transform import AffineTransform

    tform = AffineTransform(matrix=transform)
    transformed_mask: np.ndarray = affine_transform(
        mask, tform.params, output_shape=output_shape, mode="constant", order=0
    )
    return transformed_mask


def write_masks_as_hdf5(
    path: PathLike,
    name: str,
    mask: np.ndarray,
    shapes_data: dict[str, np.ndarray | str],
    display_name: str | None = None,
    creation_date: datetime | None = None,
    color: np.ndarray | None = None,
    metadata: dict[str, np.ndarray] | None = None,
) -> None:
    """Write masks to file."""
    import h5py

    # need the following attributes to be valid
    assert mask.max() <= 1, "Expected binary mask."
    assert mask.ndim == 2, "Expected 2D mask."
    # turn into a binary mask
    mask = mask.astype(bool)
    shape = mask.shape

    if display_name is None:
        display_name = name
    creation_date = creation_date or datetime.now()
    if isinstance(creation_date, datetime):
        creation_date = creation_date.strftime(TIME_FORMAT)
    if color is None:
        color = np.asarray([1.0, 0.0, 0.0, 1.0])
    if shapes_data is None:
        shapes_data = {}
    if shapes_data:
        assert "shape_data" in shapes_data, "Expected 'shape_data' key in shapes_data."
        assert "shape_types" in shapes_data, "Expected 'shape_types' key in shapes_data."

    # write to file
    with h5py.File(path, "w") as f:
        # write attributes first
        grp = f.create_group("Mask")
        grp.attrs["shape"] = tuple(shape)
        grp.attrs["name"] = name
        grp.attrs["display_name"] = display_name
        grp.attrs["creation_date"] = creation_date
        # write mask data
        grp.create_dataset("mask", data=mask)
        grp.create_dataset("color", data=color)
        # write shapes data
        if shapes_data:
            for index, (data, shape_type) in enumerate(zip(shapes_data["shape_data"], shapes_data["shape_types"])):
                grp = f.create_group(f"Mask/Shapes/{index}")
                grp.attrs["shape_type"] = shape_type
                grp.create_dataset("data", data=data)
        # add extra data
        if metadata:
            grp = f.create_group("Mask/Metadata")
            for meta_name, meta_data in metadata.items():
                if meta_data.ndim != mask.ndim or meta_data.shape != mask.shape:
                    logger.warning(
                        f"Skipped metadata key={meta_name} as it had wrong shape or dimension. {meta_data.shape}"
                    )
                    continue
                grp.create_dataset(f"{meta_name}", data=meta_data)


def write_masks_as_geojson(path: PathLike, shapes_data: dict[str, np.ndarray | str], display_name: str) -> None:
    """Write masks to GeoJSON."""
    from koyo.json import write_json_data
    from shapely import Polygon, to_geojson

    polygons = []
    for index, array in enumerate(shapes_data["shape_data"]):
        polygons.append(
            {
                "type": "Feature",
                "id": index,
                "geometry": to_geojson(Polygon(array)),
                "properties": {"classification": {"name": display_name}},
            }
        )
    write_json_data(path, polygons)


def write_masks_as_image(path: PathLike, mask: np.ndarray) -> None:
    """Write masks to image."""
    from PIL import Image

    # need the following attributes to be valid
    assert mask.max() <= 1, "Expected binary mask."
    assert mask.ndim == 2, "Expected 2D mask."
    # turn into a binary mask
    mask = mask.astype(bool)

    # write to file
    img = Image.fromarray(mask)
    img.save(path)


def is_polygon_valid(x: np.ndarray, y: np.ndarray | None = None) -> bool:
    """Check whether the polygons are valid."""
    p = Polygon(x) if y is None else Polygon((x, y))
    return p.is_valid and p.is_simple  # type: ignore[no-any-return]


def remove_invalid(
    shape_data: list[dict[str, np.ndarray | str]],
) -> tuple[list[int], list[dict[str, np.ndarray | str]]]:
    """Remove invalid polygons."""
    keep, valid_shapes = [], []
    for index, shape in enumerate(shape_data):
        if is_polygon_valid(shape["array"], None):
            valid_shapes.append(shape)
            keep.append(index)
    logger.warning(f"Removed {len(shape_data) - len(valid_shapes)} invalid polygons.")
    return keep, valid_shapes


def transform_masks(
    transform_to: PathLike,
    masks: list[PathLike],
    output_dir: PathLike,
    fmt: str | MaskOutputFmt | list[str] | list[MaskOutputFmt],
    config_path: PathLike,
    scene_index: int | None = None,
    overwrite: bool = False,
) -> None:
    """Transform and export masks."""
    from image2image_io.readers import ShapesReader, get_simple_reader
    from image2image_io.utils.warp import get_affine_from_config

    if isinstance(fmt, str):
        fmt = [fmt]
    if not fmt:
        raise ValueError("No output format specified.")

    affine, mask_shape, pixel_size = get_affine_from_config(config_path, yx=True, px=True, inv=False)
    mask_inv_pixel_size = 1 / pixel_size
    affine = np.asarray(affine, dtype=float)
    if affine.shape != (3, 3):
        raise ValueError("Expected 3x3 affine matrix.")

    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    # image that masks are transformed to
    image_reader = get_simple_reader(transform_to, init_pyramid=False, auto_pyramid=False, scene_index=scene_index)
    # mask_shape = image_reader.image_shape
    # mask_inv_pixel_size = image_reader.inv_resolution
    logger.trace(f"Initial mask shape {mask_shape} with {1 / mask_inv_pixel_size:.4f} resolution.")

    with_index = any(f in ["hdf5"] for f in fmt)
    with_shapes = any(f in ["hdf5", "geojson"] for f in fmt)
    for mask in tqdm(masks, desc="Exporting masks..."):
        with MeasureTimer() as timer:
            mask_reader = ShapesReader(mask)
            display_name = mask_reader.display_name or mask_reader.path.stem
            logger.trace(f"Reading mask {display_name} in {timer()}")

            mask_indexed = None
            mask = mask_reader.to_mask(mask_shape, inv_pixel_size=mask_inv_pixel_size, silent=False)
            if with_index:
                mask_indexed = mask_reader.to_mask(
                    mask_shape, inv_pixel_size=mask_inv_pixel_size, with_index=True, silent=False
                )
            shapes = None
            if with_shapes:
                _, shapes = mask_reader.to_shapes()
            # masks must be transformed to the image shape - sometimes that might involve warping if affine matrix
            # is specified
            transformed_mask = image_reader.warp(mask, affine=affine)
            transformed_mask_indexed = None
            if mask_indexed is not None:
                transformed_mask_indexed = image_reader.warp(mask_indexed, affine=affine)
            logger.trace(f"Transformed mask {display_name} in {timer(since_last=True)}")

            for fmt_ in fmt:
                extension = {"hdf5": "h5", "binary": "png", "geojson": "geojson"}[fmt_]
                output_path = output_dir / f"{display_name}_ds={image_reader.path.stem}.{extension}"
                if fmt_ == "hdf5":
                    write_masks_as_hdf5(
                        output_path,
                        display_name,
                        transformed_mask,
                        shapes,
                        display_name,
                        metadata={"polygon_index": transformed_mask_indexed},
                    )
                elif fmt_ == "binary":
                    write_masks_as_image(output_path, transformed_mask)
                elif fmt_ == "geojson":
                    write_masks_as_geojson(output_path, shapes, display_name)
                else:
                    raise ValueError(f"Unsupported format '{fmt_}'")
                logger.info(f"Exported {output_path} in {timer(since_last=True)}")


def transform_shapes_or_points(
    files: list[PathLike],
    output_dir: PathLike,
    config_path: PathLike,
    overwrite: bool = False,
):
    """Transform and export shapes or points."""
    from image2image_io.readers import get_simple_reader
    from image2image_io.utils.warp import ImageWarper

    transform_seq = ImageWarper(config_path, inv=False)
    for file in tqdm(files, desc="Exporting masks..."):
        with MeasureTimer() as timer:
            reader = get_simple_reader(file)
            display_name = reader.display_name
            logger.trace(f"Reading shape/point {display_name} in {timer()}")
            if reader.reader_type == "points":
                x, y = reader.x, reader.y
                transform_seq.transform_points(x, y)



"""CZI reader."""

from __future__ import annotations

import typing as ty

import numpy as np
import zarr.storage
from koyo.timer import MeasureTimer
from koyo.typing import PathLike
from loguru import logger

from image2image_io.config import CONFIG
from image2image_io.readers._base_reader import BaseReader
from image2image_io.readers._czi import CziFile, CziSceneFile, get_czi_thumbnail
from image2image_io.utils.utilities import guess_rgb

logger = logger.bind(src="CZI")


class CziMixin:
    """Mixin class for CZI images."""

    fh: CziFile | CziSceneFile
    array_shape: tuple[int, ...]
    is_rgb: bool
    _is_rgb: bool | None = None
    auto_pyramid: bool | None = None
    n_channels: int

    def _get_channel_names(self) -> list[str]:
        """Return list of channel names."""
        czi_meta = self.xml_metadata()
        channels_meta = czi_meta["ImageDocument"]["Metadata"]["DisplaySetting"]["Channels"]["Channel"]
        channel_names: list[str] = []
        for ch in channels_meta:
            if isinstance(ch, dict):
                channel_names.append(str(ch.get("ShortName")))
            else:
                channel_names.append(str(ch))
        if len(channel_names) != self.n_channels:
            # logger.warning(
            #     f"Number of channels ({self.n_channels}) does not match number of channel names ({channel_names})"
            # )
            channel_names = []
            if self.is_rgb:
                channel_names = ["R", "G", "B"]
            else:
                for idx, _ch in enumerate(range(self.n_channels)):
                    channel_names.append(f"C{str(idx + 1).zfill(2)}")
        return channel_names

    def _get_pixel_size(self) -> float:
        czi_meta = self.xml_metadata()
        pixel_scaling_str = czi_meta["ImageDocument"]["Metadata"]["Scaling"]["Items"]["Distance"][0]["Value"]
        return float(pixel_scaling_str) * 1_000_000

    def xml_metadata(self) -> dict[str, ty.Any]:
        """Return XML metadata."""
        from image2image_io.utils.utilities import xmlstr_to_dict

        # from tifffile import xml2dict
        return xmlstr_to_dict(self.fh.metadata())

    def _get_image_info(self) -> tuple:
        # if RGB need to get 0
        if self.fh.shape[-1] > 1:
            ch_dim_idx = self.fh.axes.index("0")
        else:
            ch_dim_idx = self.fh.axes.index("C")
        y_dim_idx = self.fh.axes.index("Y")
        x_dim_idx = self.fh.axes.index("X")
        shape = np.array(self.fh.shape)
        if shape[-1] > 1:
            array_shape = np.array(shape)[[y_dim_idx, x_dim_idx, ch_dim_idx]]
        else:
            array_shape = np.array(shape)[[ch_dim_idx, y_dim_idx, x_dim_idx]]
        im_shape = (shape[y_dim_idx], shape[x_dim_idx])
        return ch_dim_idx, y_dim_idx, x_dim_idx, array_shape, im_shape, self.fh.dtype


class CziImageReader(BaseReader, CziMixin):  # type: ignore[misc]
    """CZI file wrapper."""

    fh: CziFile

    def __init__(
        self,
        path: PathLike,
        key: str | None = None,
        init_pyramid: bool | None = None,
        auto_pyramid: bool | None = None,
    ):
        super().__init__(path, key, auto_pyramid=auto_pyramid)
        self.fh = CziFile(self.path)
        self.n_scenes = CziSceneFile.get_num_scenes(self.path)

        *_, self.shape, self._image_shape, self._array_dtype = self._get_image_info()
        self._is_rgb = guess_rgb(self.shape)
        self.resolution = self._get_pixel_size()
        self._channel_names = self._get_channel_names()
        CONFIG.trace(
            f"{path}: RGB={self.is_rgb}; dims={self.shape}; px={self.resolution:.3f}; n_ch={len(self._channel_names)}"
        )

        init_pyramid = init_pyramid if init_pyramid is not None else CONFIG.init_pyramid
        if init_pyramid:
            with MeasureTimer() as timer:
                self._pyramid = self.pyramid
            CONFIG.trace(f"{path}: pyramid={len(self._pyramid)} in {timer()}")

    @property
    def n_channels(self) -> int:
        """Return number of channels."""
        if self.is_rgb:
            return self.shape[2]
        _, n_channels = self.get_channel_axis_and_n_channels()
        return n_channels

    def get_dask_pyr(self) -> list:
        """Get instance of Dask pyramid."""
        auto_pyramid = self.auto_pyramid if self.auto_pyramid is not None else CONFIG.auto_pyramid
        self._zstore = zarr.storage.TempStore()
        return self.fh.zarr_pyramidize_czi(self._zstore, auto_pyramid)

    def get_thumbnail(self) -> tuple[np.ndarray, tuple[float, float]]:
        """Get thumbnail."""
        thumbnail, scale = get_czi_thumbnail(self.fh, self.scale)
        if thumbnail is None:
            return self.pyramid[-1], self.scale_for_pyramid(-1)
        return thumbnail, scale


class CziSceneImageReader(BaseReader, CziMixin):  # type: ignore[misc]
    """Multi-scene reader."""

    fh: CziSceneFile

    def __init__(
        self,
        path: PathLike,
        key: str | None = None,
        scene_index: int = 0,
        init_pyramid: bool | None = None,
        auto_pyramid: bool | None = None,
    ):
        super().__init__(path, key, reader_kws={"scene_index": scene_index}, auto_pyramid=auto_pyramid)
        self.fh = CziSceneFile(self.path, scene_index=scene_index)

        *_, self.shape, self._image_shape, self._array_dtype = self._get_image_info()
        self._is_rgb = self.fh.is_rgb
        self.resolution = self._get_pixel_size()
        self._channel_names = self._get_channel_names()
        CONFIG.trace(
            f"{path}: RGB={self.is_rgb}; dims={self.shape}; px={self.resolution:.3f}; n_ch={len(self._channel_names)}"
            f"; scene={scene_index}"
        )

        init_pyramid = init_pyramid if init_pyramid is not None else CONFIG.init_pyramid
        if init_pyramid:
            with MeasureTimer() as timer:
                self._pyramid = self.pyramid
            CONFIG.trace(f"{path}: pyramid={len(self._pyramid)} in {timer()}")

    @property
    def n_channels(self) -> int:
        """Return number of channels."""
        if self.is_rgb:
            return self.shape[2]
        _, n_channels = self.get_channel_axis_and_n_channels()
        return n_channels

    def get_dask_pyr(self) -> list:
        """Get instance of Dask pyramid."""
        auto_pyramid = self.auto_pyramid if self.auto_pyramid is not None else CONFIG.auto_pyramid
        self._zstore = zarr.storage.TempStore()
        return self.fh.zarr_pyramidize_czi(self._zstore, auto_pyramid)



"""Test points reader and utilities."""

import pandas as pd
import pytest

from image2image_io.readers.points_reader import read_points_from_df


@pytest.mark.parametrize("columns", [["x", "y"], ["x_location", "y_location"], ["x", "y_location"]])
def test_read_points_from_df(columns):
    x_col, y_col = columns
    df = pd.DataFrame({x_col: [1, 2, 3], y_col: [4, 5, 6]})
    x, y, _ = read_points_from_df(df)
    assert len(x) == 3, "Expected to read points from DataFrame."



"""Base model."""

import typing as ty
from pathlib import Path

from koyo.typing import PathLike
from pydantic import BaseModel as _BaseModel


class BaseModel(_BaseModel):
    """Base model."""

    class Config:
        """Config."""

        arbitrary_types_allowed = True

    def update(self, **kwargs: ty.Dict) -> None:
        """Update transformation."""
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def to_dict(self) -> ty.Dict:
        """Convert to dict."""
        raise NotImplementedError("Must implement method")

    def to_json(self, path: PathLike) -> None:
        """Export data as JSON."""
        from koyo.json import write_json_data

        path = Path(path)
        write_json_data(path, self.to_dict())

    @classmethod
    def from_json(cls, path: PathLike) -> "BaseModel":
        """Create from JSON."""
        from koyo.json import read_json_data

        path = Path(path)
        return cls.from_dict(read_json_data(path))

    def to_toml(self, path: PathLike) -> None:
        """Export data as TOML."""
        from koyo.toml import write_toml_data

        path = Path(path)
        write_toml_data(path, self.to_dict())

    @classmethod
    def from_toml(cls, path: PathLike) -> "BaseModel":
        """Create from TOML."""
        from koyo.toml import read_toml_data

        path = Path(path)
        return cls.from_dict(read_toml_data(path))

    @classmethod
    def from_dict(cls, data: dict) -> "BaseModel":
        """Create from dict."""
        raise NotImplementedError("Must implement method")

    @classmethod
    def from_file(cls, path: PathLike) -> "BaseModel":
        """Create from file."""
        path = Path(path)
        if path.suffix == ".json":
            return cls.from_json(path)
        elif path.suffix == ".toml":
            return cls.from_toml(path)
        raise ValueError(f"Unknown file format: '{path.suffix}'")



"""GeoJSON reader for image2image."""

from __future__ import annotations

import typing as ty
from pathlib import Path

import numpy as np
import pandas as pd
from koyo.rand import temporary_seed
from koyo.typing import PathLike
from loguru import logger
from tqdm import tqdm

from image2image_io.config import CONFIG
from image2image_io.readers._base_reader import BaseReader
from image2image_io.readers.geojson_utils import get_int_dtype, read_geojson, shape_reader
from image2image_io.readers.utilities import check_df_columns, get_column_name

if ty.TYPE_CHECKING:
    from shapely.geometry import Polygon


PATH_IF_COUNT = 1_000


def is_txt_and_has_columns(
    path: PathLike, required: list[str], either: list[tuple[str, ...]], either_dtype: tuple[np.dtype, ...] | None = None
) -> bool:
    """Check if a text file has the required columns."""
    import pandas as pd

    path = Path(path)
    if path.suffix == ".csv":
        df = pd.read_csv(path, nrows=1)
    elif path.suffix in [".txt", ".tsv"]:
        temp = pd.read_csv(path, delimiter="\t", nrows=1)
        sep = "\t" if len(temp.columns) > 1 else " "
        df = pd.read_csv(path, delimiter=sep, nrows=1)
    elif path.suffix == ".parquet":
        df = pd.read_parquet(path)
    else:
        raise ValueError(f"Invalid file extension: {path.suffix}")
    return check_df_columns(df, required, either, either_dtype)


def get_shape_columns(path: PathLike) -> tuple[str, str, str | None]:
    """Get columns."""
    path = Path(path)
    if path.suffix == ".csv":
        df = pd.read_csv(path, nrows=1)
    elif path.suffix in [".txt", ".tsv"]:
        temp = pd.read_csv(path, delimiter="\t", nrows=1)
        sep = "\t" if len(temp.columns) > 1 else " "
        df = pd.read_csv(path, delimiter=sep, nrows=1)
    elif path.suffix == ".parquet":
        df = pd.read_parquet(path)
    else:
        raise ValueError(f"Invalid file extension: {path.suffix}")
    x_key = get_column_name(df, ["vertex_x", "x", "x:x", "x_location", "x_centroid"])
    y_key = get_column_name(df, ["vertex_y", "y", "y:y", "y_location", "y_centroid"])
    group_by = get_column_name(df, ["cell", "cell_id", "shape", "shape_name"])
    return x_key, y_key, group_by


def read_shapes(path: PathLike) -> tuple:
    """Read shapes."""
    path = Path(path)
    if path.suffix == ".csv":
        df = pd.read_csv(path)
    elif path.suffix in [".txt", ".tsv"]:
        temp = pd.read_csv(path, delimiter="\t", nrows=1)
        sep = "\t" if len(temp.columns) > 1 else " "
        df = pd.read_csv(path, delimiter=sep)
    elif path.suffix == ".parquet":
        df = pd.read_parquet(path)
    else:
        raise ValueError(f"Invalid file extension: {path.suffix}")
    return read_shapes_from_df(df)


def read_shapes_from_df(df: pd.DataFrame) -> tuple:
    """Read shapes from DataFrame."""
    x_key = get_column_name(df, ["vertex_x", "x"])
    y_key = get_column_name(df, ["vertex_y", "y"])
    group_by = get_column_name(df, ["cell", "cell_id", "shape", "shape_name"])
    is_points = x_key == "x"
    shapes_geojson, shape_data = [], []
    for group, indices in df.groupby(group_by).groups.items():
        dff = df.iloc[indices]
        shape_data.append(
            {
                "array": np.c_[dff[x_key].values, dff[y_key].values].astype(np.float32),
                "shape_type": "polygon",
                "shape_name": group,
            }
        )
        shapes_geojson.append({"geometry": {"type": "Polygon"}, "properties": {"classification": {"name": group}}})
    return shapes_geojson, shape_data, is_points


def napari_to_shapes_data(name: str, data: list[np.ndarray], shapes: list[str]) -> dict:
    """Convert napari shapes to data."""
    shape_data = []
    for array, _shape_type in zip(data, shapes):
        shape_data.append(
            {
                "array": array[:, [1, 0]],
                "shape_type": "polygon",  # shape_type,
                "shape_name": name,
            }
        )
    return shape_data


def read_data(path: Path) -> dict[str, dict[str, np.ndarray | str], bool]:
    """Read data."""
    if path.suffix.lower() in [".json", ".geojson"]:
        geojson_data, shape_data, is_points = read_geojson(path)
    else:
        geojson_data, shape_data, is_points = read_shapes(path)
    return geojson_data, shape_data, is_points


class ShapesReader(BaseReader):
    """GeoJSON reader for image2image."""

    _is_rgb: bool = False
    reader_type = "shapes"
    _channel_names: list[str]
    _array_dtype = np.dtype("float32")

    def __init__(self, path: PathLike, key: str | None = None, auto_pyramid: bool | None = None, init: bool = True):
        super().__init__(path, key=key, auto_pyramid=auto_pyramid)
        if not init:
            self.geojson_data, self.shape_data = [], []
            return
        self.geojson_data, self.shape_data, is_points = read_data(self.path)
        self.display_type = "points" if is_points else None
        self._channel_names = [self.path.stem]

    @classmethod
    def create(cls, name: str = "", channel_names: list[str] | None = None):
        """Create empty instance."""
        obj = cls(name, key=name, init=False)
        if channel_names:
            obj._channel_names = channel_names
        return obj

    def remove_invalid(self) -> None:
        """Remove invalid shapes."""
        from image2image_io.utils.mask import remove_invalid

        keep, self.shape_data = remove_invalid(self.shape_data)
        self.geojson_data = [self.geojson_data[i] for i in keep]

    def to_mask(
        self, output_shape: tuple[int, int], with_index: bool = False, inv_pixel_size: float = 1.0, silent: bool = True
    ) -> np.ndarray:
        """Convert to mask.

        It's possible that GeoJSON/shapes data is in the physical units, so we need to convert it to pixels. In order to
        do so, it's necessary to multiply the coordinates by 1/pixel_size.
        """
        from image2image_io.utils.mask import polygons_to_mask, shapes_to_polygons

        # if native resolution is 1 (not set) and the specified pixel_size is not 1, then we need to multiply the
        if self.resolution != 1.0 and inv_pixel_size != 1.0:
            inv_pixel_size = 1.0

        polygons = shapes_to_polygons(
            self.shape_data, with_index=with_index, inv_pixel_size=inv_pixel_size, silent=silent
        )
        mask = polygons_to_mask(polygons, output_shape)
        return mask

    def to_mask_alt(
        self, output_size: tuple[int, int], with_index: bool = False, inv_pixel_size: float = 1.0, silent: bool = True
    ) -> np.ndarray:
        """
        Draw a binary or label mask using shape data.

        Parameters
        ----------
        output_size: tuple of int
            Size of the mask in tuple(x,y)
        with_index: bool
            Whether to write each mask instance as a label (1-n_shapes) or to write all as binary (255)
        inv_pixel_size: float
            Inverse pixel size for the mask
        silent: bool
            Whether to display progress bar

        Returns
        -------
        mask: np.ndarray
            Drawn mask at set output size

        """
        import cv2

        dtype = np.uint8
        if with_index:
            dtype = get_int_dtype(len(self.shape_data))  # type: ignore[assignment]
        mask = np.zeros(output_size[::-1], dtype=dtype)
        shapes = self.shape_data
        for idx, sh in enumerate(tqdm(shapes, desc="Drawing shapes", leave=False, miniters=1000, disable=silent)):
            shape = sh["array"]
            yx = shape * inv_pixel_size
            mask = cv2.fillPoly(
                mask,
                pts=[yx.astype(np.int32)],
                color=idx + 1 if with_index else np.iinfo(dtype).max,
            )
        return mask

    def parse_data(self) -> tuple:
        """Parse data."""
        shape_data = self.shape_data
        shapes_geojson, _, _ = shape_reader(shape_data)
        n_shapes = len(shapes_geojson)
        if n_shapes > 10_000 and CONFIG.subsample and CONFIG.subsample_ratio < 1.0:
            n_subsample = int(CONFIG.subsample_ratio * n_shapes)
            logger.trace(f"Subsampling to {n_subsample:,} shapes.")
            with temporary_seed(CONFIG.subsample_random_seed, skip_if_negative_one=True):
                indices = np.random.choice(n_shapes, n_subsample, replace=False)
            shapes_geojson = [shapes_geojson[i] for i in indices]
            shape_data = [shape_data[i] for i in indices]

        n_shapes = len(shapes_geojson)
        shape_types = [sh["geometry"]["type"] for sh in shapes_geojson]
        shape_names = [sh["properties"]["classification"]["name"] for sh in shapes_geojson]
        shape_arrays = [s["array"][:, [1, 0]] for s in shape_data]  # expect y, x
        shape_props = {"name": shape_names}
        shape_text = {
            "string": "{name}",
            "color": "white",
            "anchor": "center",
            "size": 12,
            "visible": False,
        }
        return n_shapes, shape_types, shape_names, shape_arrays, shape_props, shape_text

    def to_shapes(self) -> tuple[str, dict[str, np.ndarray | str]]:
        """Convert to shapes that can be exported to Shapes layer."""
        _, shape_types, shape_names, shape_arrays, *_ = self.parse_data()
        if len(shape_types) > PATH_IF_COUNT:
            shape_types = ["path"] * len(shape_types)
        else:
            shape_types = [s.lower() for s in shape_types]  # expected polygon not Polygon
        return shape_names[0], {"shape_types": shape_types, "shape_data": shape_arrays}

    def to_shapes_kwargs(self, edge_color: str = "cyan", **kwargs: ty.Any) -> dict:
        """Return data so it's compatible with Shapes layer."""
        n_shapes, _, _, shape_arrays, shape_props, shape_text = self.parse_data()
        if CONFIG.shape_display == "polygon or path":
            shape_type = "polygon" if n_shapes < PATH_IF_COUNT else "path"
        else:
            shape_type = CONFIG.shape_display

        kws = {
            "data": [(shape, shape_type) for shape in shape_arrays],
            "properties": shape_props,
            "text": shape_text,
            "scale": self.scale,
            "affine": self.transform,
            "edge_width": 10 if n_shapes < 500 else 3,
            "edge_color": edge_color,
        }
        kws.update(kwargs)
        return kws

    def to_points_kwargs(self, face_color: str, **kwargs: ty.Any) -> dict:
        """Return data so it's compatible with Shapes layer."""
        df = _convert_geojson_to_df(self.shape_data)
        x = df["x"].values
        y = df["y"].values
        n = len(x)
        size = 15
        if n > 5_000:
            size = 5
        elif n > 50_000:
            size = 1

        kws = {
            "data": np.c_[y, x],
            "scale": self.scale,
            "affine": self.transform,
            "face_color": face_color,
            "size": size,
        }
        kws.update(kwargs)
        return kws

    def get_dask_pyr(self) -> list[ty.Any]:
        """Get dask representation of the pyramid."""
        raise NotImplementedError("Must implement method")

    def to_shapely(self) -> list[Polygon]:
        """Convert to shapely."""
        from shapely.geometry import Polygon

        return [Polygon(s["array"]) for s in self.shape_data]

    def to_table(self):
        """Convert to table."""
        data = []
        # iterate over shapes
        for i, shape in enumerate(self.shape_data):
            name = shape["shape_name"] + f"-{i}"
            # create numpy array with x, y, shape-name columns
            data.append(np.c_[shape["array"], np.full(shape["array"].shape[0], name)])
        # concatenate all arrays
        data = np.concatenate(data)
        # create DataFrame
        df = pd.DataFrame(data, columns=["x", "y", "shape"])
        df = df.astype({"x": np.float32, "y": np.float32})
        return df

    def to_csv(self, path: PathLike, as_px: bool = False) -> Path:
        """Export data as CSV file."""
        df = self.to_table()
        df.to_csv(path, index=False)
        return Path(path)

    @property
    def shape(self) -> tuple[int, ...]:
        """Return shape of data."""
        if hasattr(self, "shape_data"):
            return (len(self.shape_data),)
        return (0,)


def _convert_geojson_to_df(shape_data: list[dict]) -> pd.DataFrame:
    """Convert GeoJSON data so that it can be transformed back to GeoJSON."""
    # types: pt = Point; pg = Polygon; mp = MultiPolygon
    data = []  # columns: x, y, unique_index, inner, outer, type
    for feature in shape_data:
        data.append(feature["array"])
    if data:
        data = np.concatenate(data)
    df = pd.DataFrame(data, columns=["x", "y"])
    return df



"""CLI Transform."""

import os

import pytest
from koyo.utilities import is_installed

from image2image_io.utils._test import get_test_file, get_test_files

has_reg = is_installed("image2image_reg")


def test_cli_entrypoint():
    """Test CLI entrypoint."""
    exit_status = os.system("i2io --help")
    assert exit_status == 0, "Exit status was not 0"


@pytest.mark.xfail(reason="needs to be resolved.")
@pytest.mark.parametrize("input_key", ["-i", "--input"])
@pytest.mark.parametrize("with_title", ["-t", "-T"])
def test_thumbnail(tmp_path, input_key, with_title):
    tmp = tmp_path
    tiffs = get_test_files("*.ome.tiff")
    files = " ".join([f"{input_key} {tiff!s}" for tiff in tiffs])
    exit_status = os.system(f"i2io thumbnail {files} -o {tmp!s} {with_title}")
    assert exit_status == 0, "Exit status was not 0"
    assert len(list(tmp.glob("*.jpg"))) > 0, "No thumbnail images"


@pytest.mark.xfail(reason="needs to be resolved.")
@pytest.mark.skipif(not has_reg, reason="image2image-reg is not installed.")
@pytest.mark.parametrize("transform", get_test_files("transform/*.i2r.json"))
def test_transform(tmp_path, transform):
    tmp = tmp_path
    tiff = get_test_file("transform/test-1.ome.tiff")
    exit_status = os.system(f"i2io transform image -i {tiff!s} -o {tmp!s} -T {transform}")
    assert exit_status == 0, "Exit status was not 0"



"""Numpy array wrapper."""

from __future__ import annotations

import numpy as np
from koyo.typing import PathLike

from image2image_io.readers._base_reader import BaseReader


class ArrayImageReader(BaseReader):
    """Reader for data that has defined coordinates."""

    is_fixed: bool = False

    def __init__(
        self,
        path: PathLike,
        array: np.ndarray,
        key: str | None = None,
        resolution: float = 1.0,
        auto_pyramid: bool | None = None,
        channel_names: list[str] | None = None,
    ):
        super().__init__(path, key, auto_pyramid=auto_pyramid)
        self.array = array
        self.resolution = resolution
        if channel_names is not None:
            self._channel_names = channel_names

    @property
    def channel_names(self) -> list[str]:
        """List of channel names."""
        if self._channel_names is not None:
            return self._channel_names
        if self.array.ndim == 2:
            return ["C0"]
        _, n_channels = self.get_channel_axis_and_n_channels()
        self._channel_names = [f"C{i}" for i in range(n_channels)]
        return self._channel_names

    @property
    def pyramid(self) -> list:
        """Pyramid."""
        return [self.array]



"""Merge images."""

from __future__ import annotations

from pathlib import Path

from koyo.timer import MeasureTimer
from koyo.typing import PathLike
from loguru import logger


def merge(
    name: str,
    paths: list[PathLike],
    output_dir: PathLike,
    fmt: str = "ome-tiff",
    as_uint8: bool | None = None,
    channel_ids: list[int] | None = None,
    overwrite: bool = False,
) -> Path:
    """Merge multiple images."""
    from image2image_io.models.merge import MergeImages
    from image2image_io.readers import get_simple_reader
    from image2image_io.writers.merge_tiff_writer import MergeOmeTiffWriter
    # from image2image_reg.wrapper import ImageWrapper

    paths = [Path(path) for path in paths]
    output_dir = Path(output_dir)

    reader_names = []
    pixel_sizes = []
    channel_names = []
    image_shapes = []
    with MeasureTimer() as timer:
        for path_ in paths:
            path = Path(path_)
            reader = get_simple_reader(path)
            reader_names.append(reader.name)
            pixel_sizes.append(reader.resolution)
            channel_names.append(reader.channel_names)
            image_shapes.append(reader.image_shape)
    logger.info(f"Loaded {len(reader_names)} images in {timer()}.")
    output_path = output_dir / f"{name}_merged-registered"
    merge_obj = MergeImages(paths, pixel_sizes, channel_names=channel_names)
    writer = MergeOmeTiffWriter(merge_obj)
    with MeasureTimer() as timer:
        writer.merge_write_image_by_plane(
            output_path.name,
            reader_names,
            output_dir=output_dir,
            as_uint8=as_uint8,
            channel_ids=channel_ids,
        )
    logger.info(f"Merged images in {timer()}.")
    return path



"""Normalizer model class."""

import numpy as np
from pydantic import field_validator

from image2image_io.models.base import BaseModel


class PreProcessor(BaseModel):
    name: str = ""

    def __call__(self, channel_name: str, image: np.ndarray) -> np.ndarray:
        """Normalize image."""
        raise NotImplementedError("Must implement method")


class NoopProcessor(PreProcessor):
    """A container class for no operation."""

    name: str = "noop"

    def __call__(self, channel_name: str, image: np.ndarray) -> np.ndarray:
        """Normalize image."""
        return image


class IMSNormalizer(PreProcessor):
    """A container class to normalize images."""

    # Attributes
    name: str = ""
    array: np.ndarray

    # noinspection PyMethodParameters
    @field_validator("array")
    def _validate_array(cls, array: np.ndarray) -> np.ndarray:
        """Validate array."""
        if not isinstance(array, np.ndarray):
            raise ValueError("Array must be a numpy array.")
        if not array.ndim == 2:
            raise ValueError("Array must be 2D.")
        return array

    def __call__(self, channel_name: str, image: np.ndarray) -> np.ndarray:
        """Normalize image."""
        if image.ndim != 2:
            raise ValueError("Image must be 2D.")
        if image.shape != self.array.shape:
            raise ValueError("Image shape does not match array")
        return image / self.array  # type: ignore[no-any-return]



"""Transform."""

import typing as ty
from copy import deepcopy
from pathlib import Path

import numpy as np
from koyo.typing import PathLike
from loguru import logger
from pydantic import PrivateAttr
from skimage.transform import AffineTransform, ProjectiveTransform

from image2image_io.enums import DEFAULT_TRANSFORM_NAME
from image2image_io.models.base import BaseModel


class TransformData(BaseModel):
    """Transformation data."""

    # Transformation object
    name: str = DEFAULT_TRANSFORM_NAME
    _transform: ty.Optional[ProjectiveTransform] = PrivateAttr(None)
    # this value should never change
    fixed_resolution: float = 1.0
    # this value can change
    moving_resolution: float = 1.0
    # Arrays of fixed and moving points
    fixed_points: ty.Optional[np.ndarray] = None
    moving_points: ty.Optional[np.ndarray] = None
    # affine transformation matrix
    affine: ty.Optional[np.ndarray] = None
    # Type of transformation
    transformation_type: str = "affine"
    # Inverse
    is_inverse: bool = False

    def to_dict(self) -> ty.Dict:
        """Serialize data."""
        return {
            "fixed_points": self.fixed_points.tolist() if self.fixed_points is not None else [],
            "moving_points": self.moving_points.tolist() if self.moving_points is not None else [],
            "fixed_pixel_size_um": self.fixed_resolution,
            "moving_pixel_size_um": self.moving_resolution,
            "matrix_yx_um": self.compute().params.tolist(),
            "matrix_yx_px": self.compute(px=True).params.tolist(),
        }

    @property
    def transform(self) -> ProjectiveTransform:
        """Retrieve the transformation object."""
        if self._transform is None:
            if self.fixed_points is None or self.moving_points is None:
                if self.affine is None:
                    self.affine = np.eye(3)
                self._transform = AffineTransform(matrix=self.affine)
            else:
                self._transform = self.compute()
        return self._transform

    def __call__(self, coords: np.ndarray) -> np.ndarray:
        """Transform coordinates."""
        if self.transform is None:
            raise ValueError("No transformation found.")
        return self.transform(coords)  # type: ignore[no-any-return]

    def inverse(self, coords: np.ndarray) -> np.ndarray:
        """Inverse transformation of coordinates."""
        if self.transform is None:
            raise ValueError("No transformation found.")
        return self.transform.inverse(coords)  # type: ignore[no-any-return]

    @property
    def matrix(self) -> np.ndarray:
        """Retrieve the transformation array."""
        if self.transform is None:
            raise ValueError("No transformation found.")
        return self.transform.params  # type: ignore[no-any-return]

    def compute(
        self, moving_resolution: ty.Optional[float] = None, yx: bool = True, px: bool = False
    ) -> ProjectiveTransform:
        """Compute transformation matrix."""
        from image2image_io.utils.utilities import compute_transform

        if moving_resolution is None:
            moving_resolution = self.moving_resolution

        moving_points = self.moving_points
        fixed_points = self.fixed_points
        if moving_points is None or len(moving_points) == 0 or fixed_points is None or len(fixed_points) == 0:
            if self.affine is None:
                self.affine = np.eye(3)
                logger.trace("Transform has not been specified - using identity matrix.")
            affine = self.affine
            return AffineTransform(matrix=affine)

        # swap yx to xy
        if not yx:
            moving_points = moving_points[:, ::-1]
            fixed_points = fixed_points[:, ::-1]
        # convert to pixels
        if not px:
            moving_points = moving_points * moving_resolution
            fixed_points = fixed_points * self.fixed_resolution
        if self.is_inverse:
            moving_points, fixed_points = fixed_points, moving_points

        # compute transform
        transform = compute_transform(
            moving_points,  # source
            fixed_points,  # destination
            self.transformation_type,
        )
        self.moving_resolution = moving_resolution
        return transform

    @classmethod
    def from_array(cls, matrix: np.ndarray) -> "TransformData":
        """Load from array."""
        matrix = np.asarray(matrix)
        assert matrix.shape == (3, 3), "Expected (3, 3) matrix"
        return TransformData(affine=matrix)


class TransformModel(BaseModel):
    """Model containing transformation data."""

    transforms: ty.Optional[ty.Dict[Path, TransformData]] = None

    class Config:
        """Config."""

        arbitrary_types_allowed = True

    @property
    def name_to_path_map(self) -> ty.Dict[ty.Union[str, Path], Path]:
        """Returns dictionary that maps transform name to path."""
        if self.transforms is None:
            return {}

        mapping: ty.Dict[PathLike, Path] = {}
        for name in self.transforms:
            if isinstance(name, str):
                name = Path(name)
            mapping[name.name] = name
            mapping[Path(name.name)] = name
            mapping[name] = name
        return mapping

    @property
    def transform_names(self) -> ty.List[str]:
        """Return list of transform names."""
        return [Path(t).name for t in self.transforms] if self.transforms else []

    def add_transform(
        self, name_or_path: PathLike, transform_data: TransformData, with_inverse: bool = True, silent: bool = False
    ) -> None:
        """Add transformation matrix."""
        if self.transforms is None:
            self.transforms = {}

        path = Path(name_or_path)
        self.transforms[path] = transform_data
        if not silent:
            logger.info(f"Added '{path.name}' to list of transformations")
        if with_inverse:
            path = Path(name_or_path).parent / (path.name + " (inverse)")
            transform_data = deepcopy(transform_data)
            transform_data.is_inverse = True
            self.transforms[path] = transform_data
            if not silent:
                logger.info(f"Added '{path.name}' to list of transformations")

    def remove_transform(self, name_or_path: PathLike) -> None:
        """Remove transformation matrix."""
        if self.transforms is None:
            return

        def _remove_transform(path_of_transform: Path) -> None:
            if path_of_transform in self.transforms:
                del self.transforms[path_of_transform]
                logger.info(f"Removed '{path_of_transform.name}' from list of transformations")
            else:
                for path in self.transforms:
                    if path.name == path_of_transform.name:
                        del self.transforms[path]
                        logger.info(f"Removed '{path.name}' from list of transformations")
                        break

        name_or_path = Path(name_or_path)
        _remove_transform(name_or_path)
        inv_name_or_path = name_or_path.parent / (name_or_path.name + " (inverse)")
        _remove_transform(inv_name_or_path)

    def get_matrix(self, name_or_path: PathLike) -> ty.Optional[TransformData]:
        """Get transformation matrix."""
        if self.transforms is None:
            return None

        name_or_path = Path(name_or_path)
        name_or_path = self.name_to_path_map.get(name_or_path, None)  # type: ignore
        if name_or_path is None:
            return None
        if name_or_path in self.transforms:
            return self.transforms[name_or_path]
        return None

    def get_info(self) -> str:
        info = ""
        transform = self.compute()
        if hasattr(transform, "scale"):
            scale = transform.scale
            scale = (scale, scale) if isinstance(scale, float) else scale
            info += f"\nScale: {scale[0]:.3f}, {scale[1]:.3f}"
        if hasattr(transform, "translation"):
            translation = transform.translation
            translation = (translation, translation) if isinstance(translation, float) else translation
            info += f"\nTranslation: {translation[0]:.3f}, {translation[1]:.3f}"
        if hasattr(transform, "rotation"):
            radians = transform.rotation
            degrees = radians * 180 / 3.141592653589793
            info += f"\nRotation: {radians:.3f} ({degrees:.3f}°)"
        return info



"""Geo-JSON reader."""

from __future__ import annotations

import json
import typing as ty
import zipfile
from pathlib import Path

import cv2
import geojson
import numpy as np
from koyo.typing import PathLike
from shapely.geometry import Polygon

GJ_SHAPE_TYPE = {
    "polygon": geojson.Polygon,
    "multipolygon": geojson.MultiPolygon,
    "point": geojson.Point,
    "multipoint": geojson.MultiPoint,
    "multilinestring": geojson.MultiLineString,
    "linestring": geojson.LineString,
}
GJ_SHAPE_TYPE_NAME = {
    "polygon": "Polygon",
    "multipolygon": "MultiPolygon",
    "point": "Point",
    "multipoint": "MultiPoint",
    "multilinestring": "MultiLineString",
    "linestring": "LineString",
}


def get_int_dtype(value: int) -> np.dtype:
    """Determine appropriate bit precision for indexed image.

    Parameters
    ----------
    value:int
        number of shapes

    Returns
    -------
    dtype:np.dtype
        Appropriate data type for the number of masks.
    """
    if value <= np.iinfo(np.uint8).max:
        return np.uint8
    if value <= np.iinfo(np.uint16).max:
        return np.uint16
    if value <= np.iinfo(np.uint32).max:
        return np.int32
    else:
        raise ValueError("Too many shapes")


def geojson_to_numpy(gj: dict) -> tuple[list[dict], bool]:
    """
    Convert geojson representation to np.ndarray representation of shape.

    Parameters
    ----------
    gj : dict
        GeoJSON data stored as python dict

    Returns
    -------
    dict
        containing keys
            "array": np.ndarray - x,y point data in array
            "shape_type": str - indicates GeoJSON shape_type (Polygon, MultiPolygon, etc.)
            "shape_name": str - name inherited from QuPath GeoJSON
    """
    is_points = False
    if gj["geometry"].get("type") == "MultiPolygon":
        pts = []
        for geo in gj["geometry"].get("coordinates"):
            for geo_ in geo:
                geo_ = np.squeeze(geo_)
                pts.append(np.asarray(np.asarray(Polygon(geo_).exterior.coords)))
    elif gj["geometry"].get("type") == "Polygon":
        try:
            coordinates = np.squeeze(np.asarray(gj["geometry"].get("coordinates")))
            pts = np.asarray(Polygon(coordinates).exterior.coords)
        except (ValueError, TypeError):
            pts = []
            for poly in gj["geometry"].get("coordinates"):
                pts.append(np.asarray(Polygon(np.asarray(np.squeeze(poly))).exterior.coords))
        # pts = np.squeeze(np.asarray(gj["geometry"].get("coordinates")))
    elif gj["geometry"].get("type") == "Point":
        pts = np.expand_dims(np.asarray(gj["geometry"].get("coordinates")), 0)
        is_points = True
    elif gj["geometry"].get("type") == "MultiPoint":
        pts = np.asarray(gj["geometry"].get("coordinates"))
        is_points = True
    elif gj["geometry"].get("type") == "LineString":
        pts = np.asarray(gj["geometry"].get("coordinates"))
    else:
        raise ValueError(f"GeoJSON type {gj['geometry'].get('type')} not supported")

    if "properties" not in gj:
        shape_name = "unnamed"
    else:
        if not gj["properties"].get("classification"):
            shape_name = "unnamed"
        else:
            shape_name = gj["properties"].get("classification").get("name")

    if isinstance(pts, list):
        return [
            {
                "array": np.asarray(pt).astype(np.float32),
                "shape_type": gj["geometry"].get("type"),
                "shape_name": shape_name,
            }
            for pt in pts
        ], is_points
    # elif len(pts.shape) == 1:
    #     return [
    #         {
    #             "array": np.asarray(pts[0]).astype(np.float32),
    #             "shape_type": gj["geometry"].get("type"),
    #             "shape_name": shape_name,
    #         }
    #     ]
    else:
        return [
            {
                "array": pts.astype(np.float32),
                "shape_type": gj["geometry"].get("type"),
                "shape_name": shape_name,
            }
        ], is_points


def add_unnamed(gj: dict) -> dict:
    """Add unnamed object."""
    if "properties" not in gj:
        gj["properties"] = {"classification": {"name": "unnamed"}}
    else:
        if not gj["properties"].get("classification"):
            gj["properties"].update({"classification": {"name": "unnamed"}})
    return gj


def read_geojson(json_file: PathLike) -> tuple[list, list, bool]:
    """Read GeoJSON files (and some QuPath metadata).

    Parameters
    ----------
    json_file : str
        file path of QuPath exported GeoJSON

    Returns
    -------
    gj_data : list
        GeoJSON information
    shapes_np : list
        GeoJSON information stored in np.ndarray
            "array": np.ndarray - x,y point data in array
            "shape_type": str - indicates GeoJSON shape_type (Polygon, MultiPolygon, etc.)
            "shape_name": str - name inherited from QuPath GeoJSON
    """
    if Path(json_file).suffix != ".zip":
        gj_data = json.load(open(json_file))
    else:
        with zipfile.ZipFile(json_file, "r") as z:
            for filename in z.namelist():
                with z.open(filename) as f:
                    data = f.read()
                    gj_data = json.loads(data.decode("utf-8"))
    return _parse_geojson_data(gj_data)


def _parse_geojson_data(gj_data: dict | list) -> tuple[list, list, bool]:
    if isinstance(gj_data, dict):
        # handle GeoPandas GeoJSON
        if "type" in gj_data and "features" in gj_data:
            gj_data = gj_data["features"]
        else:
            gj_data = [gj_data]

    shapes_np, is_points = [], False
    for gj in gj_data:
        data, is_points = geojson_to_numpy(gj)
        shapes_np.extend(data)
    gj_data = [add_unnamed(gj) for gj in gj_data]
    return gj_data, shapes_np, is_points


def numpy_to_geojson(
    np_array: np.ndarray, shape_type: str = "polygon", shape_name: str = "unnamed"
) -> tuple[dict, dict]:
    """Convert np.ndarray to GeoJSON dict.

    Parameters
    ----------
    np_array: np.ndarray
        coordinates of data
    shape_type:str
        GeoJSON shape type
    shape_name:str
        Name of the shape

    Returns
    -------
    shape_gj : dict
        GeoJSON information
    shape_np : dict
        GeoJSON information stored in np.ndarray
            "array": np.ndarray - x,y point data in array
            "shape_type": str - indicates GeoJSON shape_type (Polygon, MultiPolygon, etc.)
            "shape_name": str - name inherited from QuPath GeoJSON
    """
    sh_type = shape_type.lower()

    gj_func = GJ_SHAPE_TYPE[sh_type]
    if sh_type == "polygon":
        np_array = np.vstack([np_array, np_array[0, :]])
        geometry = gj_func([np_array.tolist()])
    elif sh_type in ["multipoint", "linestring"]:
        geometry = gj_func(np_array.transpose().tolist())
    else:
        geometry = gj_func(np_array.tolist())

    shape_gj = {
        "type": "Feature",
        "id": "annotation",
        "geometry": geometry,
        "properties": {
            "classification": {"name": shape_name, "colorRGB": -1},
            "isLocked": False,
        },
    }

    shape_np = {
        "array": np_array,
        "shape_type": shape_type.lower(),
        "shape_name": shape_name,
    }
    return shape_gj, shape_np


def shape_reader(shape_data: list, **kwargs: ty.Any) -> tuple[dict, dict, bool]:
    """Read shape data for transformation.

    Shape data is stored as numpy arrays for operations but also as GeoJSON
    to contain metadata and interface with QuPath.

    Parameters
    ----------
    shape_data: list of np.ndarray or str
        if str, will read data as GeoJSON file, if np.ndarray with assume
        it is coordinates
    kwargs
        keyword args passed to np_to_geojson convert

    Returns
    -------
    shapes_gj: list of dicts
        list of dicts of GeoJSON information
    shapes_np: list of dicts
        list of dicts of GeoJSON information stored in np.ndarray
        "array": np.ndarray - x,y point data in array
        "shape_type": str - indicates GeoJSON shape_type (Polygon, MultiPolygon, etc.)
        "shape_name": str - name inherited from QuPath GeoJSON
    """
    if not isinstance(shape_data, list):
        shape_data = [shape_data]

    shapes_gj = []
    shapes_np = []
    is_points = False
    for sh in shape_data:
        if isinstance(sh, dict):
            out_shape_gj, out_shape_np = numpy_to_geojson(sh["array"], sh["shape_type"], sh["shape_name"])

        elif isinstance(sh, np.ndarray):
            out_shape_gj, out_shape_np = numpy_to_geojson(sh, **kwargs)

        else:
            if Path(sh).is_file():
                sh_fp = Path(sh)

                if sh_fp.suffix in [".json", ".geojson", ".zip"]:
                    out_shape_gj, out_shape_np, is_points = read_geojson(str(sh_fp))
                # elif sh_fp.suffix == ".cz":
                #     out_shape_gj = read_zen_shapes(str(sh_fp))
                #     out_shape_np = [gj_to_np(s) for s in out_shape_gj]
                else:
                    raise ValueError(f"{sh_fp!s} is not a geojson or numpy array")
            else:
                raise FileNotFoundError(f"{Path(sh).as_posix()!s} file not found")

        if isinstance(out_shape_gj, list):
            shapes_gj.extend(out_shape_gj)
        else:
            shapes_gj.append(out_shape_gj)

        if isinstance(out_shape_np, list):
            shapes_np.extend(out_shape_np)
        else:
            shapes_np.append(out_shape_np)
    return shapes_gj, shapes_np, is_points


def scale_shape_coordinates(poly: dict, scale_factor: float):
    """
    Scale coordinates by a factor.

    Parameters
    ----------
    poly: dict
        dict of coordinate data contain np.ndarray in "array" key
    scale_factor: float
        isotropic scaling factor for the coordinates

    Returns
    -------
    poly: dict
        dict containing coordinates scaled by scale_factor
    """
    poly_coords = poly["array"]
    poly_coords = poly_coords * scale_factor
    poly["array"] = poly_coords
    return poly


def invert_nonrigid_transforms(itk_transforms: list):
    """Check list of sequential ITK transforms for non-linear (i.e., bspline) transforms.

    Transformations need to be inverted to transform from moving to fixed space as transformations
    are mapped from fixed space to moving.
    This will first convert any non-linear transforms to a displacement field then invert the displacement field
    using ITK methods. It usually works quite well but is not an exact solution.
    Linear transforms can be inverted on  the fly when transforming points.

    Parameters
    ----------
    itk_transforms:list
        list of itk.Transform

    Returns
    -------
    itk_transforms:list
        list of itk.Transform where any non-linear transforms are replaced with an inverted displacement field
    """
    tform_linear = [t.is_linear for t in itk_transforms]

    if all(tform_linear):
        return itk_transforms
    else:
        nl_idxs = np.where(np.array(tform_linear) == 0)[0]
        for nl_idx in nl_idxs:
            if not itk_transforms[nl_idx].inverse_transform:
                print(
                    f"transform at index {nl_idx} is non-linear and the inverse has not been computed\n"
                    "inverting displacement field(s)...\n"
                    "this can take some time"
                )
                itk_transforms[nl_idx].compute_inverse_nonlinear()

    return itk_transforms


# def prepare_pt_transformation_data(transformations, compute_inverse=True):
#     """
#     Read and prepare wsireg transformation data for point set transformation
#
#     Parameters
#     ----------
#     transformations
#         list of dict containing elastix transformation data or str to wsireg .json file containing
#         elastix transformation data
#     compute_inverse : bool
#         whether or not to compute the inverse transformation for moving to fixed point transformations
#
#     Returns
#     -------
#     itk_pt_transforms:list
#         list of transformation data ready to operate on points
#     target_res:
#         physical spacing of the final transformation in the transform sequence
#         This is needed to map coordinates defined as pixel indices to physical coordinates and then back
#     """
#     if not all([isinstance(t, RegTransform) for t in transformations]):
#         _, transformations = wsireg_transforms_to_itk_composite(transformations)
#     if compute_inverse:
#         transformations = invert_nonrigid_transforms(transformations)
#     target_res = float(transformations[-1].output_spacing[0])
#     return transformations, target_res
#
#
# def itk_transform_pts(
#     pt_data: np.ndarray,
#     itk_transforms: list,
#     px_idx=True,
#     source_res=1,
#     output_idx=True,
#     target_res=2,
# ):
#     """
#     Transforms x,y points stored in np.ndarray using list of ITK transforms
#     All transforms are in physical coordinates, so all points must be converted to physical coordinates
#     before transformation, but this function allows converting back to pixel indices after transformation
#
#     Can intake points in physical coordinates is px_idx == False
#     Can output points in physical coordinates if output_idx == False
#
#     Parameters
#     ----------
#     pt_data : np.ndarray
#         array where rows are points and columns are x,y
#     itk_transforms: list
#         list of ITK transforms, non-linear transforms should be inverted
#     px_idx: bool
#         whether points are specified in physical coordinates (i.e., microns) or
#         in pixel indices
#     source_res: float
#         resolution of the image on which annotations were made
#     output_idx: bool
#         whether transformed points should be output in physical coordinates (i.e., microns) or
#         in pixel indices
#     target_res: float
#         resolution of the final target image for conversion back to pixel indices
#
#     Returns
#     -------
#     tformed_pts:np.ndarray
#         transformed points array where rows are points and columns are x,y
#
#     """
#     tformed_pts = []
#     for pt in pt_data:
#         if px_idx :
#             pt = pt * source_res
#         for idx, t in enumerate(itk_transforms):
#             if idx == 0:
#                 t_pt = t.inverse_transform.TransformPoint(pt)
#             else:
#                 t_pt = t.inverse_transform.TransformPoint(t_pt)
#         t_pt = np.array(t_pt)
#         if output_idx :
#             t_pt *= 1 / target_res
#         tformed_pts.append(t_pt)
#
#     return np.stack(tformed_pts)


def get_all_shape_coords(shapes: list):
    """Get all shape coordinates from a list of shapes."""
    return np.vstack([np.squeeze(sh["geometry"]["coordinates"][0]) for sh in shapes])


# code below is for managing transforms as masks rather than point sets
# will probably not reimplement, if segmentation data can is expressed
# as a mask, it can be transformed as an image (using attachment_modality)
def approx_polygon_contour(mask: np.ndarray, percent_arc_length=0.01):
    """
    Approximate binary mask contours to polygon vertices using cv2.

    Parameters
    ----------
    mask : numpy.ndarray
        2-d numpy array of datatype np.uint8.
    percent_arc_length : float
        scaling of epsilon for polygon approximate vertices accuracy.
        maximum distance of new vertices from original.

    Returns
    -------
    numpy.ndarray
        returns an 2d array of vertices, rows: points, columns: y,x

    """
    contours, hierarchy = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)

    if len(contours) > 1:
        contours = [contours[np.argmax([cnt.shape[0] for cnt in contours])]]

    epsilon = percent_arc_length * cv2.arcLength(contours[0], True)
    approx = cv2.approxPolyDP(contours[0], epsilon, True)
    return np.squeeze(approx).astype(np.uint32)


# def index_mask_to_shapes(index_mask, shape_name, tf_shapes):
#     """
#     Find the polygons of a transformed shape mask, conveting binary mask
#     to list of polygon verteces and sorting by numerical index
#
#     Parameters
#     ----------
#     index_mask:np.ndarray
#         mask where each shape is defined by it's index
#     shape_name:str
#         name of the shape
#     tf_shapes:list
#         original list of shape GeoJSON data to be updated
#
#     Returns
#     -------
#     updated_shapes:list
#         dict of GeoJSON information with updated coordinate information
#     """
#     labstats = sitk.LabelShapeStatisticsImageFilter()
#     labstats.SetBackgroundValue(0)
#     labstats.Execute(index_mask)
#
#     index_mask = sitk.GetArrayFromImage(index_mask)
#     updated_shapes = deepcopy(tf_shapes)
#
#     for idx, shape in enumerate(tf_shapes):
#         if shape["properties"]["classification"]["name"] == shape_name:
#             label_bb = labstats.GetBoundingBox(idx + 1)
#             x_min = label_bb[0]
#             x_len = label_bb[2]
#             y_min = label_bb[1]
#             y_len = label_bb[3]
#
#             sub_mask = index_mask[y_min : y_min + y_len, x_min : x_min + x_len]
#
#             sub_mask[sub_mask == idx + 1] = 255
#
#             yx_coords = approx_polygon_contour(sub_mask, 0.00001)
#             xy_coords = yx_coords
#             xy_coords = np.append(xy_coords, xy_coords[:1, :], axis=0)
#             xy_coords = xy_coords + [x_min, y_min]
#             updated_shapes[idx]["geometry"]["coordinates"] = [xy_coords.tolist()]
#
#     return updated_shapes


# don't intend to maintain
# def read_zen_shapes(zen_fp):
#     """Read Zeiss Zen Blue .cz ROIs files to wsimap shapely format.
#
#     Parameters
#     ----------
#     zen_fp : str
#         file path of Zen .cz.
#
#     Returns
#     -------
#     list
#         list of wsimap shapely rois
#
#     """
#
#     root = etree.parse(zen_fp)
#
#     rois = root.xpath("//Elements")[0]
#     shapes_out = []
#     for shape in rois:
#         try:
#             ptset_name = shape.find("Attributes/Name").text
#         except AttributeError:
#             ptset_name = "unnamed"
#
#         if shape.tag == "Polygon":
#             ptset_cz = shape.find("Geometry/Points")
#             # ptset_type = "Polygon"
#
#             poly_str = ptset_cz.text
#             poly_str = poly_str.split(" ")
#             poly_str = [poly.split(",") for poly in poly_str]
#             poly_str = [[float(pt[0]), float(pt[1])] for pt in poly_str]
#
#             poly = {
#                 "geometry": geojson.Polygon(poly_str),
#                 "properties": {"classification": {"name": ptset_name}},
#             }
#
#             shapes_out.append(poly)
#
#         if shape.tag == "Rectangle":
#             rect_pts = shape.find("Geometry")
#
#             x = float(rect_pts.findtext("Left"))
#             y = float(rect_pts.findtext("Top"))
#             width = float(rect_pts.findtext("Width"))
#             height = float(rect_pts.findtext("Height"))
#
#             rect = geojson.Polygon(
#                 [
#                     [x, y],
#                     [x + width, y],
#                     [x + width, y + height],
#                     [x, y + height],
#                     [x, y],
#                 ]
#             )
#
#             rect = {
#                 "geometry": rect,
#                 "properties": {"classification": {"name": ptset_name}},
#             }
#
#             shapes_out.append(rect)
#
#     return shapes_out



from __future__ import annotations

import typing as ty
from pathlib import Path

import numpy as np
from koyo.typing import PathLike
from loguru import logger

from image2image_io.config import CONFIG
from image2image_io.readers._base_reader import BaseReader


class ImageWrapper:
    """Wrapper around image data."""

    data: dict[str, BaseReader]
    resolution: float = 1.0

    def __init__(self, reader: dict[str, BaseReader] | None = None):
        self.data = reader or {}

        resolution = [1.0]
        for _, _reader_or_array in self.data.items():
            if hasattr(_reader_or_array, "base_layer_pixel_res"):
                resolution.append(_reader_or_array.resolution)
        self.resolution = np.min(resolution)

    def __repr__(self) -> str:
        return f"ImageWrapper<{len(self.data)}>"

    def __getitem__(self, item: int | PathLike) -> BaseReader:
        if isinstance(item, int):
            return list(self.data.values())[item]
        return self.data[item]

    def add(self, reader: BaseReader) -> None:
        """Add data to wrapper."""
        self.data[reader.key] = reader
        logger.trace(f"Added '{reader.key}' to ImageWrapper.")

    def remove(self, key_or_reader: str | BaseReader) -> None:
        """Remove data from wrapper."""
        if hasattr(key_or_reader, "key"):
            key_or_reader = key_or_reader.key
        if key_or_reader in self.data:
            reader = self.data.pop(key_or_reader, None)
            del reader

    def remove_path(self, path: PathLike) -> list[str]:
        """Remove readers for specified path."""
        keys = self.get_key_for_path(path)
        for key in keys:
            self.remove(key)
        return keys

    def get_key_for_path(self, path: PathLike) -> list[str]:
        """Return key(s) for the specified path."""
        path = Path(path)
        keys = []
        for reader in self.reader_iter():
            if reader.path == path:
                keys.append(reader.key)
            elif reader.key == path.name:
                keys.append(reader.key)
        return keys

    def get_reader_for_path(self, path: PathLike) -> BaseReader:
        """Return reader for specified path."""
        keys = self.get_key_for_path(path)
        if len(keys) == 1:
            return self.data[keys[0]]
        raise ValueError(f"Reader for path '{path}' not found.")

    def get_reader_for_key(self, key: str) -> BaseReader:
        """Return reader for specified key."""
        try:
            return self.data[key]
        except KeyError:
            return None

    def is_loaded(self, path: PathLike) -> bool:
        """Check if the path is loaded."""
        from image2image_io.readers import sanitize_read_path

        path = Path(path)
        for reader in self.reader_iter():
            if reader.path == path or reader.path == sanitize_read_path(path, raise_on_error=False):
                return True
        return False

    def channel_names_for_names(self, names: ty.Sequence[str]) -> list[str]:
        """Return list of channel names for a given wrapper/dataset."""
        clean_names = []
        for name in names:
            if isinstance(name, Path):
                name = str(name.name)
            if "| " not in name:
                name = f"| {name}"
            clean_names.append(name)

        channel_names = []
        for channel_name in self.channel_names():
            for name in clean_names:
                if channel_name.endswith(name):
                    channel_names.append(channel_name)
        return channel_names

    def map_channel_to_index(self, dataset: str, channel_name: str) -> int:
        """Map channel name to index."""
        dataset_to_channel_map: dict[str, list[str]] = {}
        for name in self.channel_names():
            dataset, channel = name.split(" | ")
            dataset_to_channel_map.setdefault(dataset, []).append(channel)
        channels: list[str] = dataset_to_channel_map[dataset]
        return channels.index(channel_name)

    def channel_image_iter(self) -> ty.Generator[tuple[str, list[np.ndarray]], None, None]:
        """Iterator of channel name + image."""
        yield from zip(self.channel_names(), self.image_iter())

    def channel_image_for_channel_names_iter(
        self, channel_names: list[str] | None
    ) -> ty.Generator[tuple[str, list[np.ndarray] | None, BaseReader], None, None]:
        """Iterate of channel name + image for a specified list of channels."""
        if channel_names is None:
            channel_names = self.channel_names()
        for channel_name in channel_names:
            name, dataset = channel_name.split(" | ")
            reader = self.data[dataset]
            index = reader.channel_to_index(name)
            if reader.reader_type == "image":
                yield channel_name, reader.get_channel_pyramid(index), reader
            else:
                yield channel_name, None, reader

    def channel_image_reader_iter(self) -> ty.Generator[tuple[str, list[np.ndarray], BaseReader], None, None]:
        """Iterator of channel name + image."""
        for channel_name, (_, reader, image, _) in zip(self.channel_names(), self.reader_data_iter()):
            yield channel_name, image, reader

    def path_reader_iter(self) -> ty.Generator[tuple[Path, BaseReader], None, None]:
        """Iterator of a path + reader."""
        for reader in self.reader_iter():
            yield reader.path, reader

    def reader_channel_iter(self) -> ty.Generator[tuple[str, BaseReader, int], None, None]:
        """Iterator to add channels."""
        for reader_name, reader in self.data.items():
            yield from self.reader_channel_iter_for_reader(reader_name, reader)

    def reader_channel_iter_for_reader(
        self, reader_name: str, reader: BaseReader
    ) -> ty.Generator[tuple[str, BaseReader, int], None, None]:
        """Iterator to add channels."""
        if reader.reader_type == "shapes":
            yield reader_name, reader, 0
        elif reader.reader_type == "points":
            yield reader_name, reader, 0
        else:
            # spetial case for RGB images
            if reader.is_rgb and not CONFIG.split_rgb:
                yield reader_name, reader, None
            else:
                if reader.channel_names:
                    for index, _ in enumerate(reader.channel_names):
                        yield reader_name, reader, index
                else:
                    yield from self._reader_channel_iter(reader_name, reader)
                # for reader_name_, index in self._reader_channel_iter(reader_name, reader):
                #     # for reader_name_, _, _, index in self._reader_image_iter(reader_name, reader):
                #     yield reader_name_, reader, index

    @staticmethod
    def _reader_channel_iter(
        reader_name: str, reader: BaseReader
    ) -> ty.Generator[tuple[str, BaseReader, int], None, None]:
        """Iterator to add channels."""
        for channel_index in range(reader.n_channels):
            yield reader_name, reader, channel_index

    def reader_data_iter(
        self,
    ) -> ty.Generator[tuple[str, BaseReader, list[np.ndarray] | None, int], None, None]:
        """Iterator to add channels."""
        for reader_name, reader_or_array in self.data.items():
            if reader_or_array.reader_type == "shapes":
                yield reader_name, reader_or_array, None, 0
            elif reader_or_array.reader_type == "points":
                yield reader_name, reader_or_array, None, 0
            else:
                yield from self._reader_image_iter(reader_name, reader_or_array)

    def reader_image_iter(
        self,
    ) -> ty.Generator[tuple[str, BaseReader, list[np.ndarray], int], None, None]:
        """Iterator to add channels."""
        for reader_name, reader_or_array in self.data.items():
            yield from self._reader_image_iter(reader_name, reader_or_array)

    @staticmethod
    def _reader_image_iter(
        reader_name: str, reader: BaseReader
    ) -> ty.Generator[tuple[str, BaseReader, list[np.ndarray], int], None, None]:
        # image is a numpy array
        temp = reader.pyramid
        array = temp if isinstance(temp, list) else [temp]

        # get the shape of the 'largest image in pyramid'
        shape = reader.image_shape
        # get the number of dimensions, which determines how images are split into channels
        ndim = len(shape)
        channel_axis, n_channels = reader.get_channel_axis_and_n_channels()

        for channel_index in range(n_channels):
            # 2D image or is RGB image
            if channel_axis is None:
                yield reader_name, reader, array, channel_index
            # 3D image where the first axis corresponds to different channels
            elif channel_axis == 0:
                yield reader_name, reader, [a[channel_index] for a in array], channel_index
            # 3D image where the second axis corresponds to different channels
            elif channel_axis == 1:
                yield reader_name, reader, [a[:, channel_index] for a in array], channel_index
            # 3D image where the last axis corresponds to different channels
            elif channel_axis == 2:
                yield reader_name, reader, [a[..., channel_index] for a in array], channel_index
            else:
                raise ValueError(f"Cannot read image with {ndim} dimensions")

    def image_iter(self) -> ty.Iterator[list[np.ndarray]]:
        """Iterator to add channels."""
        for _, _, image, _ in self.reader_image_iter():
            yield image

    def reader_iter(self) -> ty.Iterator[BaseReader]:
        """Iterator of readers."""
        for _, reader_or_array in self.data.items():
            yield reader_or_array

    def dataset_names(self, reader_type: tuple[str, ...] = ("all",)) -> list[str]:
        """Return list of channel names."""
        names = []
        show_all = "all" in reader_type
        for key, reader in self.data.items():
            if show_all or reader.reader_type in reader_type:
                names.append(key)
        return names

    def channel_names(self) -> list[str]:
        """Return list of channel names."""
        names = []
        for key, reader, index in self.reader_channel_iter():
            channel_names = self.get_channel_names_for_reader(key, reader, index)
            names.extend(channel_names)
        return names

    @staticmethod
    def get_channel_names_for_reader(key: str, reader: BaseReader, index: int) -> list[str]:
        """Retrieve channel names for specified reader."""
        try:
            if reader.is_rgb and not CONFIG.split_rgb:
                channel_names = ["RGB"]
            else:
                channel_names = [reader.channel_names[index]]
        except (IndexError, NotImplementedError):
            channel_names = [f"C{index}"]
        channel_names = [f"{name} | {key}" for name in channel_names]
        return channel_names

    @property
    def min_resolution(self) -> float:
        """Return minimum resolution."""
        resolutions = [reader.resolution for reader in self.reader_iter()]
        if resolutions:
            return min(resolutions)
        return 1.0

    @staticmethod
    def get_affine(reader: BaseReader, moving_resolution: float) -> np.ndarray:
        """Get affine transformation for specified reader."""
        transform_data = reader.transform_data
        if not transform_data:
            raise ValueError("No transformation data found")
        affine = transform_data.compute(moving_resolution=moving_resolution, px=False)
        return np.asarray(affine.params)



"""CZI file reader."""

from __future__ import annotations

import typing as ty
import warnings
from concurrent.futures import ThreadPoolExecutor
from functools import cached_property
from multiprocessing import cpu_count
from pathlib import Path
from xml.etree import ElementTree

import dask.array as da
import numpy as np
import zarr
from czifile import CziFile as _CziFile
from czifile import DimensionEntryDV1, DirectoryEntryDV
from koyo.timer import MeasureTimer
from loguru import logger
from tifffile import create_output
from tqdm import tqdm

from image2image_io.config import CONFIG
from image2image_io.readers.utilities import compute_sub_res, guess_rgb

logger = logger.bind(src="CZI")


class CziFile(_CziFile):
    """Sub-class of CziFile with added functionality to only read certain channels."""

    def asarray_alt(self, resize=True, order=0, out=None, max_workers=None):
        """Return image data from file(s) as numpy array.

        Parameters
        ----------
        resize : bool
            If True (default), resize sub/supersampled subblock data.
        order : int
            The order of spline interpolation used to resize sub/supersampled
            subblock data. Default is 0 (nearest neighbor).
        out : numpy.ndarray, str, or file-like object; optional
            Buffer where image data will be saved.
            If numpy.ndarray, a writable array of compatible dtype and shape.
            If str or open file, the file name or file object used to
            create a memory-map to an array stored in a binary file on disk.
        max_workers : int
            Maximum number of threads to read and decode subblock data.
            By default up to half the CPU cores are used.

        """
        if out is None:
            out = create_output(out, self.shape, self.dtype)

        if max_workers is None:
            max_workers = cpu_count() // 2

        def func(directory_entry, resize=resize, order=order, start=self.start, out=out):
            """Read, decode, and copy subblock data."""
            subblock = directory_entry.data_segment()
            tile = subblock.data(resize=resize, order=order)
            index = tuple(slice(i - j, i - j + k) for i, j, k in zip(directory_entry.start, start, tile.shape))
            try:
                out[index] = tile
            except ValueError as e:
                warnings.warn(str(e), stacklevel=2)
            with pbar.get_lock():
                pbar.update()

        if max_workers > 1:
            self._fh.lock = True
            with tqdm(
                total=len(self.filtered_subblock_directory), desc="Reading subblocks", mininterval=0.5, unit="block"
            ) as pbar:
                with ThreadPoolExecutor(max_workers) as executor:
                    executor.map(func, self.filtered_subblock_directory)
            self._fh.lock = None
        else:
            with tqdm(
                total=len(self.filtered_subblock_directory), desc="Reading subblocks", mininterval=0.5, unit="block"
            ) as pbar:
                for directory_entry in self.filtered_subblock_directory:
                    func(directory_entry)

        if hasattr(out, "flush"):
            out.flush()
        return out

    def zarr_pyramidize_czi(self, zarr_fp, pyramid: bool = True, tile_size: int = 1024) -> list:
        """Create a pyramidal zarr store from a CZI file."""
        dask_pyr = []
        root = zarr.open_group(zarr_fp, mode="a")
        root.attrs["axes_names"] = list(self.axes)
        root.attrs["orig_shape"] = list(self.shape)

        all_axes = list(self.axes)
        yx_dims = np.where(np.isin(all_axes, ["Y", "X"]) == 1)[0].tolist()
        yx_shape = np.array(self.shape[slice(yx_dims[0], yx_dims[1] + 1)])

        pyramid_seq = 0
        rgb_chunk = self.shape[-1] if self.shape[-1] > 2 else 1
        chunking = (1, 1, 1, tile_size, tile_size, rgb_chunk)
        out_shape = list(self.shape)
        out_dtype = self.dtype
        synchronizer = zarr.ThreadSynchronizer() if CONFIG.multicore else None
        out = root.create_dataset(
            pyramid_seq,
            shape=tuple(out_shape),
            chunks=chunking,
            dtype=out_dtype,
            overwrite=True,
            synchronizer=synchronizer,
        )

        with MeasureTimer() as timer:
            self.as_tzcyx0_array(out=out, max_workers=cpu_count() if CONFIG.multicore else 1)
            z_array = da.squeeze(da.from_zarr(zarr.open(zarr_fp)[0]))
            dask_pyr.append(z_array)
        logger.trace(f"Loaded data in {timer()} to shape {z_array.shape}")

        if not pyramid:
            logger.trace("Pyramid creation disabled")
            return dask_pyr

        ds = 1
        while np.min(yx_shape) // 2**ds >= tile_size:
            ds += 1

        ds_levels = list(range(1, ds))
        logger.trace(f"Generating {ds_levels} down-sampled images")
        for ds_factor in ds_levels:
            with MeasureTimer() as timer:
                zres = zarr.storage.TempStore()
                rgb_chunk = self.shape[-1] if self.shape[-1] > 2 else 1
                is_rgb = True if rgb_chunk > 1 else False
                z_array_ds = compute_sub_res(z_array, ds_factor, tile_size, is_rgb, self.dtype)
                da.to_zarr(z_array_ds, zres, component="0")
                dask_pyr.append(da.squeeze(da.from_zarr(zres, component="0")))
            logger.trace(f"Down-sampled {ds_factor} in {timer()} to shape {z_array_ds.shape}")
        return dask_pyr

    @cached_property
    def pos_x_um(self) -> float:
        """Return the position of the image in micrometers."""
        return self.scale_x_um * min((dim_entry.start for dim_entry in self._iter_dim_entries("X")), default=0.0)

    @cached_property
    def pos_y_um(self) -> float:
        """Return the position of the image in micrometers."""
        return self.scale_y_um * min((dim_entry.start for dim_entry in self._iter_dim_entries("Y")), default=0.0)

    @cached_property
    def pos_z_um(self) -> float:
        """Return the position of the image in micrometers."""
        return self.scale_z_um * min((dim_entry.start for dim_entry in self._iter_dim_entries("Z")), default=0.0)

    @cached_property
    def pos_t_seconds(self) -> float:
        """Return the position of the image in seconds."""
        return self.scale_t_seconds * min((dim_entry.start for dim_entry in self._iter_dim_entries("T")), default=0.0)

    @cached_property
    def scale_x_um(self) -> float:
        """Return the scale of the image in micrometers."""
        return self._get_scale("X", multiplier=10.0**6)

    @cached_property
    def scale_y_um(self) -> float:
        """Return the scale of the image in micrometers."""
        return self._get_scale("Y", multiplier=10.0**6)

    @cached_property
    def scale_z_um(self) -> float:
        """Return the scale of the image in micrometers."""
        return self._get_scale("Z", multiplier=10.0**6)

    @cached_property
    def scale_t_seconds(self) -> float:
        """Return the scale of the image in seconds."""
        return self._get_scale("T")

    @cached_property
    def channel_names(self) -> list[str] | None:
        """Return the names of the channels."""
        if "C" in self.axes:
            channel_elements = self._metadata_xml.findall(".//Metadata/Information/Image/Dimensions/Channels/Channel")
            if len(channel_elements) == self.shape[self.axes.index("C")]:
                return [c.attrib.get("Name", c.attrib["Id"]) for c in channel_elements]
        return None

    @cached_property
    def is_rgb(self) -> bool:
        """Return True if the image is RGB."""
        return "0" in self.axes and self.shape[self.axes.index("0")] > 1

    def as_tzcyx0_array(self, *args, **kwargs) -> np.ndarray:
        """Return image data as numpy array with axes TZCYX0."""
        data = self.asarray_alt(*args, **kwargs)
        tzcyx0_axis_indices = []
        if "T" in self.axes:
            tzcyx0_axis_indices.append(self.axes.index("T"))
        else:
            tzcyx0_axis_indices.append(data.ndim)
            data = np.expand_dims(data, -1)
        if "Z" in self.axes:
            tzcyx0_axis_indices.append(self.axes.index("Z"))
        else:
            tzcyx0_axis_indices.append(data.ndim)
            data = np.expand_dims(data, -1)
        if "C" in self.axes:
            tzcyx0_axis_indices.append(self.axes.index("C"))
        else:
            tzcyx0_axis_indices.append(data.ndim)
            data = np.expand_dims(data, -1)
        tzcyx0_axis_indices.append(self.axes.index("Y"))
        tzcyx0_axis_indices.append(self.axes.index("X"))
        if "0" in self.axes:
            tzcyx0_axis_indices.append(self.axes.index("0"))
        else:
            tzcyx0_axis_indices.append(data.ndim)
            data = np.expand_dims(data, -1)
        for axis_index in range(len(self.axes)):
            if axis_index not in tzcyx0_axis_indices:
                tzcyx0_axis_indices.append(axis_index)
        data = data.transpose(tzcyx0_axis_indices)
        data.shape = data.shape[:6]
        return data

    def _get_scale(self, dimension: str, multiplier: float = 1.0):
        scale_element = self._metadata_xml.find(f'.//Metadata/Scaling/Items/Distance[@Id="{dimension}"]/Value')
        if scale_element is not None:
            scale = float(scale_element.text)
            if scale > 0:
                return scale * multiplier
        return 1.0

    @cached_property
    def _metadata_xml(self) -> ElementTree.Element:
        return ElementTree.fromstring(self.metadata())

    def _iter_dim_entries(self, dimension: str) -> ty.Iterable[DimensionEntryDV1]:
        for dir_entry in self.filtered_subblock_directory:
            for dim_entry in dir_entry.dimension_entries:
                if dim_entry.dimension == dimension:
                    yield dim_entry


class CziSceneFile(CziFile):
    @staticmethod
    def get_num_scenes(path: str | Path, *args, **kwargs) -> int:
        """Get number of scenes."""
        with _CziFile(path, *args, **kwargs) as czi_file:
            if "S" in czi_file.axes:
                return czi_file.shape[czi_file.axes.index("S")]
            return 1

    def __init__(self, path: str | Path, scene_index: int, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(str(path), *args, **kwargs)
        self.scene_index = scene_index

    @cached_property
    def filtered_subblock_directory(self) -> list[DirectoryEntryDV]:
        dir_entries = super().filtered_subblock_directory
        return list(
            filter(
                lambda dir_entry: self._get_scene_index(dir_entry) == self.scene_index,
                dir_entries,
            )
        )

    @staticmethod
    def _get_scene_index(dir_entry: DirectoryEntryDV) -> int:
        scene_indices = {dim_entry.start for dim_entry in dir_entry.dimension_entries if dim_entry.dimension == "S"}
        if len(scene_indices) == 0:
            return 0
        assert len(scene_indices) == 1
        return scene_indices.pop()


def czi_tile_grayscale(rgb_image):
    """
    Convert RGB image data to greyscale.

    Parameters
    ----------
    rgb_image: np.ndarray
        image data

    Returns
    -------
    image:np.ndarray
        returns 8-bit greyscale image for 24-bit RGB image
    """
    result = (
        (rgb_image[..., 0] * 0.2125).astype(np.uint8)
        + (rgb_image[..., 1] * 0.7154).astype(np.uint8)
        + (rgb_image[..., 2] * 0.0721).astype(np.uint8)
    )

    return np.expand_dims(result, axis=-1)


def get_level_blocks(czi: CziFile) -> dict:
    """Get level blocks."""
    level_blocks: dict = {}
    for idx, sb in enumerate(czi.subblocks()):
        if sb.pyramid_type != 0:
            level = sb.shape[3] // sb.stored_shape[3]
        else:
            level = 0

        if level not in level_blocks:
            level_blocks[level] = []
        level_blocks[level].append((idx, sb))
    return level_blocks


def get_czi_thumbnail(
    czi: CziFile, pixel_spacing: tuple[int, int] | tuple[float, float]
) -> tuple[tuple[np.ndarray] | None, tuple[float, float] | None]:
    """Get CZI thumbnail."""
    ch_idx = czi.axes.index("C")
    l_blocks = get_level_blocks(czi)
    lowest_im = np.max(list(l_blocks.keys()))
    if lowest_im == 0:
        calc_thumbnail_spacing = np.asarray(pixel_spacing) * 1
    else:
        calc_thumbnail_spacing = np.asarray(pixel_spacing) * lowest_im

    thumbnail_spacing = (float(calc_thumbnail_spacing[0]), float(calc_thumbnail_spacing[1]))

    block_indices = [b[0] for b in l_blocks[lowest_im]]
    if guess_rgb(czi.shape) and len(block_indices) == 1:
        data_idx = block_indices[0]
        image_data = czi.subblock_directory[data_idx].data_segment()
        thumbnail_array = np.squeeze(image_data.data(resize=False))
        return thumbnail_array, thumbnail_spacing

    elif len(block_indices) == czi.shape[czi.axes.index("C")]:
        thumbnail_shape = list(czi.subblock_directory[block_indices[0]].data_segment().stored_shape)
        thumbnail_shape[ch_idx] = czi.shape[ch_idx]

        thumbnail_array = np.empty(thumbnail_shape, dtype=czi.dtype)
        thumbnail_array = np.squeeze(thumbnail_array)
        for b_index in block_indices:
            image_data = czi.subblock_directory[b_index].data_segment()
            data_ch_index = next(de.start for de in image_data.dimension_entries if de.dimension == "C")
            data = np.squeeze(image_data.data(resize=False))
            thumbnail_array[data_ch_index, :, :] = data
        return thumbnail_array, thumbnail_spacing
    return None, None



"""Models."""

from image2image_io.models.base import BaseModel
from image2image_io.models.merge import MergeImages
from image2image_io.models.transform import TransformData

__all__ = ["BaseModel", "MergeImages", "TransformData"]



from pathlib import Path

import appdirs

APP_NAME = "image2image"
APP_AUTHOR = False

USER_DATA_DIR = Path(appdirs.user_data_dir(APP_NAME, APP_AUTHOR))

USER_CONFIG_DIR = USER_DATA_DIR / "Configs"
USER_CONFIG_DIR.mkdir(parents=True, exist_ok=True)
USER_LOG_DIR = USER_DATA_DIR / "Logs"
USER_LOG_DIR.mkdir(parents=True, exist_ok=True)
